# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
from queue import Empty
from queue import Queue
import threading
import time

from gi.repository import Adw
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango

from komikku.consts import COVER_HEIGHT
from komikku.consts import COVER_WIDTH
from komikku.consts import DOWNLOAD_MAX_DELAY
from komikku.consts import LOGO_SIZE
from komikku.consts import MISSING_IMG_RESOURCE_PATH
from komikku.utils import convert_and_resize_image
from komikku.utils import CoverPicture
from komikku.utils import html_escape
from komikku.utils import log_error_traceback

THUMB_WIDTH = 96
THUMB_HEIGHT = 136


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/card_tracking.ui')
class TrackingDialog(Adw.PreferencesDialog):
    __gtype_name__ = 'TrackingDialog'

    group = Gtk.Template.Child('group')
    selected_tracker_row = None
    tracker_rows = []

    def __init__(self, window):
        super().__init__(follows_content_size=True)

        self.window = window

        for _id, tracker in self.window.trackers.trackers.items():
            row = TrackerRow(window, tracker)
            self.group.add(row)
            self.tracker_rows.append(row)

        self.search_subpage = TrackingSearchSubPage(self.window)
        self.search_subpage.connect('hiding', self.on_search_subpage_hiding)

        self.connect('closed', self.on_closed)

    def add_tracking(self, id):
        def run():
            manga = self.window.card.manga
            if manga.tracking is None:
                manga.tracking = {}

            tracker = self.selected_tracker_row.tracker
            data = tracker.get_manga_data(id)
            if data:
                data['_synced'] = True

                manga.tracking[tracker.id] = data

                manga.update({'tracking': manga.tracking})

            GLib.idle_add(complete, data)

        def complete(data):
            if data is None:
                self.add_toast(Adw.Toast.new(_('Failed to get data from tracker')))
                return

            self.pop_subpage()
            self.selected_tracker_row.init(data)

        self.selected_tracker_row.set_arrow_visible(True)
        self.selected_tracker_row.btn.set_visible(False)

        thread = threading.Thread(target=run)
        thread.daemon = True
        thread.start()

    def on_closed(self, _dialog):
        self.window.trackers.sync()
        self.pop_subpage()

    def on_search_subpage_hiding(self, _page):
        self.search_subpage.thread_covers_stop_flag = True
        self.search_subpage.clear()

    def show(self):
        count = 0
        for tracker_row in self.tracker_rows:
            if tracker_row.init():
                count += 1

        if count == 0:
            self.window.add_notification(_('No tracking services available'))
            return

        self.present(self.window)

    def show_search(self, tracker):
        self.push_subpage(self.search_subpage)
        self.search_subpage.init(tracker, self.window.card.manga.name)


class TracherResultRow(Gtk.ListBoxRow):
    def __init__(self, window, data):
        self.window = window
        self.data = data

        super().__init__()

        box = Gtk.Box(margin_top=6, margin_bottom=6, margin_start=6, margin_end=6, spacing=6)

        self.cover_bin = Adw.Bin()
        self.cover_bin.set_size_request(THUMB_WIDTH, THUMB_HEIGHT)
        box.append(self.cover_bin)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)

        hbox = Gtk.Box(spacing=6)

        vbox_details = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        title = Gtk.Label(
            label=html_escape(data['name']),
            use_markup=True,
            ellipsize=Pango.EllipsizeMode.END,
            xalign=0,
            hexpand=True
        )
        title.add_css_class('title')
        vbox_details.append(title)

        if data.get('authors'):
            label = Gtk.Label(
                label=html_escape(data['authors']),
                use_markup=True,
                ellipsize=Pango.EllipsizeMode.END,
                xalign=0,
                hexpand=True
            )
            label.add_css_class('subtitle')
            vbox_details.append(label)

        details = []
        if data.get('start_date'):
            details.append(data['start_date'])

        if 'status' in data:
            details.append(data['status'] or _('Unknown status'))

        if data.get('score'):
            details.append(f'{data["score"]}/10')

        label = Gtk.Label(
            label=' · '.join(details),
            use_markup=True,
            ellipsize=Pango.EllipsizeMode.END,
            xalign=0,
            hexpand=True
        )
        label.add_css_class('subtitle')
        vbox_details.append(label)

        hbox.append(vbox_details)
        vbox.append(hbox)
        box.append(vbox)

        if data.get('synopsis'):
            label = Gtk.Label(
                label=html_escape(data['synopsis'].replace('\n', '')),
                use_markup=True,
                ellipsize=Pango.EllipsizeMode.END,
                xalign=0,
                hexpand=True,
                lines=4,
                wrap=True
            )
            label.add_css_class('caption')
            vbox.append(label)

        self.btn = Gtk.Button(valign=Gtk.Align.START)
        self.btn.set_label(_('Track'))
        self.btn.connect('clicked', lambda _btn: self.window.card.tracking_dialog.add_tracking(data['id']))
        hbox.append(self.btn)

        self.set_child(box)

    def set_cover(self, data):
        picture = CoverPicture.new_from_data(data, THUMB_WIDTH, THUMB_HEIGHT, True) if data else None
        if picture is None:
            picture = CoverPicture.new_from_resource(MISSING_IMG_RESOURCE_PATH, THUMB_WIDTH, THUMB_HEIGHT)

        self.cover_bin.set_child(picture)


class TrackerRow(Adw.ExpanderRow):
    def __init__(self, window, tracker):
        self.window = window
        self.tracker = tracker

        super().__init__(title=self.tracker.name)

        self.set_logo()

        self.set_arrow_visible(False)

        self.btn = Gtk.Button(valign=Gtk.Align.CENTER)
        self.btn.set_label(_('Use'))
        self.btn.connect('clicked', self.on_btn_clicked)
        self.add_suffix(self.btn)

        # Action (tracker page, delete)
        self.action_row = Adw.ActionRow()
        cancel_button = Gtk.Button.new_from_icon_name('user-trash-symbolic')
        cancel_button.props.valign = Gtk.Align.CENTER
        cancel_button.set_tooltip_text(_('Cancel Tracking'))
        cancel_button.connect('clicked', self.cancel_tracking)
        self.action_row.add_suffix(cancel_button)
        self.add_row(self.action_row)

        # Status
        self.status_row = Adw.ComboRow(title=_('Status'))
        statuses = Gtk.StringList()
        for _status, internal_status in self.tracker.STATUSES_MAPPING.items():
            statuses.append(_(self.tracker.INTERNAL_STATUSES[internal_status]))
        self.status_row.set_model(statuses)
        self.status_changed_handler_id = self.status_row.connect('notify::selected', self.update_tracking_data)
        self.add_row(self.status_row)

        # Score
        self.score_row = Adw.SpinRow(title=_('Score'), wrap=True)
        self.score_changed_handler_id = self.score_row.connect('notify::value', self.update_tracking_data)
        self.add_row(self.score_row)

        # Chapters progress
        self.chapters_progress_row = Adw.SpinRow(title=_('Chapter'), wrap=True)
        self.num_chapter_changed_handler_id = self.chapters_progress_row.connect('notify::value', self.update_tracking_data)
        self.add_row(self.chapters_progress_row)

    def cancel_tracking(self, _btn):
        def confirm_callback():
            manga = self.window.card.manga
            manga.tracking.pop(self.tracker.id)
            manga.update({'tracking': manga.tracking})
            self.init()

        self.window.open_dialog(
            _('Delete?'),
            body=_('Are you sure you want to cancel tracking?'),
            confirm_label=_('Delete'),
            confirm_callback=confirm_callback,
            confirm_appearance=Adw.ResponseAppearance.DESTRUCTIVE
        )

    def get_logo(self):
        def run():
            try:
                res = self.tracker.save_logo()
            except Exception:
                res = False

            GLib.idle_add(complete, res)

        def complete(res):
            if not res:
                self.window.application.logger.info('Failed to get `%s` tracker logo', self.tracker.id)

            self.set_logo(use_fallback=True)

        thread = threading.Thread(target=run)
        thread.daemon = True
        thread.start()

    def init(self, data=None):
        def run():
            try:
                id = self.window.card.manga.tracking[self.tracker.id]['id']
                data = self.tracker.get_manga_data(id)
            except Exception as e:
                log_error_traceback(e)
                data = None

            if data is None:
                data = self.window.card.manga.tracking[self.tracker.id]

            GLib.idle_add(complete, data)

        def complete(data):
            if data is None:
                # Failed to get data from tracker (connection error, server error or invalid request)
                return

            self.set_enable_expansion(True)
            self.set_expanded(True)
            self.set_arrow_visible(True)
            self.set_activatable(True)
            self.set_subtitle('')
            self.btn.set_visible(False)

            if data.get('url'):
                tracker_manga_url = data['url']
            else:
                tracker_manga_url = self.tracker.get_manga_url(data['id'])
            self.action_row.set_title(f'<a href="{tracker_manga_url}">{html_escape(data["name"])}</a>')

            with self.chapters_progress_row.handler_block(self.num_chapter_changed_handler_id):
                adj = Gtk.Adjustment(
                    value=float(data['chapters_progress']) or 0,
                    lower=0,
                    upper=data['chapters'] or 10000,
                    step_increment=1,
                    page_increment=10
                )
                self.chapters_progress_row.set_adjustment(adj)

            with self.score_row.handler_block(self.score_changed_handler_id):
                self.score_row.format = self.tracker.get_user_score_format(data['score_format'])
                adj = Gtk.Adjustment(
                    value=data['score'] or 0,
                    lower=self.score_row.format['min'],
                    upper=self.score_row.format['max'],
                    step_increment=self.score_row.format['step'],
                )
                self.score_row.configure(adj, 0, self.score_row.format['step'] * 10 if self.score_row.format['step'] != 1 else 0)

            with self.status_row.handler_block(self.status_changed_handler_id):
                self.status_row.set_selected(self.tracker.get_status_index(data['status']))

        # Is the tracker granted?
        granted, access_token_valid = self.tracker.is_granted()
        self.set_visible(granted)

        if data is None:
            tracked = self.window.card.manga.tracking and self.window.card.manga.tracking.get(self.tracker.id)

            if access_token_valid and tracked:
                thread = threading.Thread(target=run)
                thread.daemon = True
                thread.start()
            else:
                self.set_expanded(False)
                self.set_enable_expansion(False)
                self.action_row.set_title('')

                if access_token_valid:
                    self.set_arrow_visible(False)
                    self.btn.set_visible(True)
                    self.set_activatable(True)
                    self.set_subtitle('')
                else:
                    self.set_arrow_visible(True)
                    self.btn.set_visible(False)
                    self.set_activatable(False)
                    self.set_subtitle(_('Connection is expired'))
        else:
            complete(data)

        return granted and access_token_valid

    def on_btn_clicked(self, _btn):
        self.window.card.tracking_dialog.show_search(self.tracker)
        self.window.card.tracking_dialog.selected_tracker_row = self

    def set_arrow_visible(self, visible):
        action_row = self.get_first_child().get_first_child().get_first_child()
        arrow_img = action_row.get_first_child().get_last_child().get_last_child()
        arrow_img.set_visible(visible)

    def set_logo(self, use_fallback=False):
        if self.tracker.logo_url:
            if self.tracker.logo_path:
                self.logo = Gtk.Image()
                self.logo.set_pixel_size(LOGO_SIZE)
                self.logo.props.margin_top = 9
                self.logo.props.margin_bottom = 9
                self.logo.set_from_file(self.tracker.logo_path)
            elif use_fallback:
                # Fallback to an Adw.Avatar if logo fetching failed
                self.logo = Adw.Avatar.new(LOGO_SIZE, self.tracker.name, True)
            else:
                self.get_logo()
                return

        else:
            self.logo = Adw.Avatar.new(LOGO_SIZE, self.tracker.name, True)

        self.add_prefix(self.logo)

    def update_tracking_data(self, _row, _gparam):
        data = {
            'chapters_progress': self.chapters_progress_row.get_value(),
            'score': self.score_row.get_value() * self.score_row.format['raw_factor'],  # RAW
            'status': self.tracker.get_status_from_index(self.status_row.get_selected()),
            '_synced': False,
        }

        manga = self.window.card.manga
        manga.tracking[self.tracker.id].update(data)
        manga.update({'tracking': manga.tracking})


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/card_tracking_search.ui')
class TrackingSearchSubPage(Adw.NavigationPage):
    __gtype_name__ = 'TrackingSearchSubPage'

    searchentry = Gtk.Template.Child('searchentry')
    stack = Gtk.Template.Child('stack')
    listbox = Gtk.Template.Child('listbox')
    no_results_status_page = Gtk.Template.Child('no_results_status_page')

    def __init__(self, window):
        super().__init__()

        self.window = window
        self.queue = Queue()
        self.thread_covers = None
        self.thread_covers_stop_flag = False

        self.tracker = None

        self.searchentry.connect('activate', lambda _w: self.search())
        self.searchentry.connect('search-changed', self.on_search_changed)

    def clear(self):
        # Empty queue
        while not self.queue.empty():
            try:
                self.queue.get()
                self.queue.task_done()
            except Empty:
                continue

        # Empty group
        row = self.listbox.get_first_child()
        while row:
            next_row = row.get_next_sibling()

            self.listbox.remove(row)
            row = next_row

    def init(self, tracker, name=None):
        self.tracker = tracker

        self.searchentry.set_placeholder_text(_('Search {}').format(self.tracker.name))
        self.searchentry.set_text(name)

        self.search()

    def on_search_changed(self, _entry):
        if not self.searchentry.get_text().strip():
            self.stack.set_visible_child_name('intro')

    def render_covers(self):
        """
        Fetch and display covers of result rows

        This method is interrupted when the navigation page is left.
        """
        def run():
            while not self.queue.empty() and self.thread_covers_stop_flag is False:
                try:
                    row = self.queue.get()
                except Empty:
                    continue
                else:
                    try:
                        data, _etag, rtime = self.tracker.get_image(row.data['cover'])
                        # Covers in landscape format are converted to portrait format
                        data = convert_and_resize_image(data, COVER_WIDTH, COVER_HEIGHT)
                    except Exception:
                        pass
                    else:
                        GLib.idle_add(row.set_cover, data)

                        if rtime:
                            time.sleep(min(2 * rtime, DOWNLOAD_MAX_DELAY))

                    self.queue.task_done()

        self.thread_covers_stop_flag = False
        self.thread_covers = threading.Thread(target=run)
        self.thread_covers.daemon = True
        self.thread_covers.start()

    def search(self):
        def run(term):
            try:
                results = self.tracker.search(term)
            except Exception as e:
                log_error_traceback(e)
                results = None

            GLib.idle_add(complete, results)

        def complete(results):
            if results:
                for result in results:
                    row = TracherResultRow(self.window, result)
                    self.listbox.append(row)
                    self.queue.put(row)

                self.stack.set_visible_child_name('results')
                self.render_covers()

            elif results is None:
                self.no_results_status_page.set_title(_('Oops, search failed. Please try again.'))
                self.stack.set_visible_child_name('no_results')

            else:
                self.no_results_status_page.set_title(_('No Results Found'))
                self.no_results_status_page.set_description(_('Try a different search'))
                self.stack.set_visible_child_name('no_results')

        self.clear()
        self.stack.set_visible_child_name('loading')

        term = self.searchentry.get_text().strip()

        self.thread = threading.Thread(target=run, args=(term,))
        self.thread.daemon = True
        self.thread.start()



# SPDX-FileCopyrightText: 2020-2026 Liliana Prikler
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Liliana Prikler <liliana.prikler@gmail.com>
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import json
import logging
import re

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import skip_past

logger = logging.getLogger(__name__)


class Dynasty(Server):
    id = 'dynasty'
    name = 'Dynasty Reader'
    lang = 'en'
    is_nsfw = True

    base_url = 'https://dynasty-scans.com'
    logo_url = base_url + '/assets/favicon-96599a954b862bfaaa71372cc403a6ab.png'
    search_url = base_url + '/search'
    manga_url = base_url + '/{0}'
    chapter_url = base_url + '/chapters/{0}'
    tags_url = base_url + '/tags/suggest/'

    filters = [
        {
            'key': 'classes',
            'type': 'select',
            'name': _('Categories'),
            'description': _('Filter by Types'),
            'value_type': 'multiple',
            'options': [
                {'key': 'Anthology', 'name': _('Anthologies'), 'default': True},
                {'key': 'Doujin', 'name': _('Doujins'), 'default': True},
                {'key': 'Issue', 'name': _('Issues'), 'default': True},
                {'key': 'Series', 'name': _('Series'), 'default': True},
                {'key': 'Chapter', 'name': _('Chapters'), 'default': False},
            ],
        },
        {
            'key': 'with_tags',
            'type': 'entry',
            'name': _('With Tags'),
            'description': _('Comma separated list of tags to search for'),
            'default': '',
        },
        {
            'key': 'without_tags',
            'type': 'entry',
            'name': _('Without Tags'),
            'description': _('Comma separated list of tags to exclude from search'),
            'default': '',
        },
    ]
    long_strip_genres = ['Long strip', ]

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'user-agent': USER_AGENT})

    @classmethod
    def get_manga_initial_data_from_url(cls, url):
        if idx := skip_past(url, 'dynasty-scans.com/'):
            return {'slug': url[idx:]}

        return None

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update({
            'authors': [],
            'scanlators': [],
            'genres': [],
            'status': None,
            'synopsis': None,
            'chapters': [],
            'server_id': self.id,
            'cover': None,
        })

        class_ = initial_data['slug'].split('/')[0]

        return {
            'anthologies': self._fill_data_multi,
            'chapters': self._fill_data_single,
            'doujins': self._fill_data_multi,
            'issues': self._fill_data_multi,
            'series': self._fill_data_multi,
        }[class_](class_, data, soup)

    def _fill_data_single(self, class_, data, soup):
        # Fill metadata
        name_element = soup.select_one('#chapter-title')
        data['name'] = name_element.b.text.strip()
        data['authors'] = [elt.text.strip() for elt in name_element.select('a')]

        details = soup.select_one('#chapter-details')
        data['scanlators'] = [elt.text.strip() for elt in details.select('span.scanlators a')]

        data['genres'] = [elt.text.strip() for elt in details.select('span.tags a')]

        date_text = details.select_one('span.released').text.strip()

        data['chapters'].append({
            'slug': data['slug'].split('/')[-1],
            'title': data['name'],
            'date': convert_date_string(date_text),
        })

        # Use first page as cover
        for script_element in soup.select('script'):
            script = script_element.string
            if script is None:
                continue

            for line in script.split('\n'):
                line = line.strip()
                if line.startswith('var pages'):
                    pages = line.replace('var pages = ', '')[:-1]
                    break
            if pages is not None:
                pages = json.loads(pages)
                break

        if pages is not None:
            data['cover'] = self.base_url + pages[0]['image']

        return data

    def _fill_data_multi(self, class_, data, soup):
        status = None

        if class_ in ('anthologies', 'series'):
            name_element = soup.select_one('.tag-title > b')
            authors_elements = soup.select('.tag-title > a')
            cover_element = soup.select_one('.cover-chapters .cover img.thumbnail')
            full_title = soup.select_one('.tag-title').text
            if len(full_title.split('—')) == 2:
                status = full_title.split('—')[1].strip()
            data['name'] = name_element.text.split('›')[-1].strip()

        elif class_ == 'doujins':
            name_element = soup.select_one('h2')
            authors_elements = soup.select('.tag-tags a')
            cover_element = soup.select_one('.tag-images a.thumbnail img')
            data['name'] = name_element.text.split('›')[-1].strip()

        elif class_ == 'issues':
            name_element = soup.select_one('.tag-title b')
            authors_elements = None
            cover_element = soup.select_one('.cover-chapters .cover img.thumbnail')
            full_title = soup.select_one('.tag-title').text
            if len(full_title.split('—')) == 2:
                status = full_title.split('—')[1].strip()
            data['name'] = name_element.text.strip()

        if cover_element:
            data['cover'] = self.base_url + cover_element.get('src')

        if authors_elements:
            data['authors'] = [element.text.strip() for element in authors_elements]

        if status == 'Ongoing':
            data['status'] = 'ongoing'
        elif status == 'Completed':
            data['status'] = 'complete'
        elif status == 'On Hiatus':
            data['status'] = 'hiatus'
        elif status in ('Abandoned', 'Cancelled', 'Dropped', 'Not Updated', 'Removed'):
            data['status'] = 'suspended'

        # Genres
        for element in soup.select('.tag-tags a.label'):
            value = element.text.strip()
            if value not in data['genres']:
                data['genres'].append(value)

        # Synopsis
        if elements := soup.select('.description p'):
            data['synopsis'] = '\n\n'.join([p.text.strip() for p in elements])

        # Chapters
        for element in soup.select('dl.chapter-list dd'):
            if a_element := element.select_one('a'):
                date_text = None
                if small := element.select_one('small:-soup-contains("release")'):
                    date_text = small.text[8:].strip()

                data['chapters'].append({
                    'slug': a_element.get('href').split('/')[-1],
                    'title': a_element.text.strip(),
                    'date': convert_date_string(date_text) if date_text else None,
                })

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        pages = None
        for script_element in soup.select('script'):
            script = script_element.string
            if script is None:
                continue

            for line in script.split('\n'):
                line = line.strip()
                if line.startswith('var pages'):
                    pages = line.replace('var pages = ', '')[:-1]
                    break
            if pages is not None:
                pages = json.loads(pages)
                break

        if pages is None:
            return None

        data = {
            'pages': [],
        }
        for page in pages:
            data['pages'].append({
                'slug': None,  # not necessary, we know image url directly
                'image': self.base_url + page['image'],
            })

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': page['image'].split('?')[0].split('/')[-1],
        }

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def resolve_tag(self, search_tag):
        r = self.session_post(self.tags_url, params={'query': search_tag})
        if r.status_code != 200:
            return None

        for tag in r.json():
            if tag['name'].lower() == search_tag.lower():
                return tag['id']

        return None

    def get_manga_list(self, term=None, classes=None, with_tags='', without_tags=''):
        with_tags = [self.resolve_tag(t.strip()) for t in with_tags.split(',') if t]
        without_tags = [self.resolve_tag(t.strip()) for t in without_tags.split(',') if t]
        classes = sorted(classes, key=str.lower) if classes else []

        params = {
            'q': '',
            'classes[]': classes,
            'with[]': with_tags,
            'without[]': without_tags,
            'sort': '',
        }

        if term:
            # Search
            params['q'] = term
        else:
            # Latest
            params['sort'] = 'created_at'

        r = self.session_get(self.search_url, params=params)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select('dl.chapter-list dd a.name'):
            slug = a_element.get('href').lstrip('/')
            if slug.startswith('chapters/'):
                # Check if chapter belongs to a serie
                if matches := re.search(r'([a-z_\/]*)_ch[0-9_]*', slug):
                    slug = matches.group(1).replace('chapters/', 'series/')

            results.append({
                'slug': slug,
                'name': a_element.text.strip(),
            })

        return results

    def get_latest_updates(self, classes=None, with_tags='', without_tags=''):
        return self.get_manga_list(classes=classes, with_tags=with_tags, without_tags=without_tags)

    def search(self, term, classes=None, with_tags='', without_tags=''):
        return self.get_manga_list(term=term, classes=classes, with_tags=with_tags, without_tags=without_tags)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import datetime
from gettext import gettext as _
import time

import requests

from komikku.consts import DOWNLOAD_MAX_DELAY
from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.utils import get_buffer_mime_type
from komikku.utils import get_response_elapsed
from komikku.utils import is_number

SEARCH_RESULTS_PAGES = 5


class Zenko(Server):
    id = 'zenko'
    name = 'Zenko'
    lang = 'uk'

    is_nsfw = True

    base_url = 'https://zenko.online'
    logo_url = base_url + '/favicon.ico'

    manga_url = base_url + '/titles/{0}'
    chapter_url = base_url + '/titles/{0}/{1}'
    image_url = 'https://zenko.b-cdn.net/{0}?optimizer=image&width=560&quality=70&height=auto'

    api_url = 'https://zenko-api.onrender.com'
    api_search_url = api_url + '/titles'
    api_manga_url = api_url + '/titles/{0}'
    api_chapters_url = api_url + '/titles/{0}/chapters'
    api_chapter_url = api_url + '/chapters/{0}'

    filters = [
        {
            'key': 'age_limit',
            'type': 'select',
            'name': _('Age Restriction'),
            'description': _('Filter by Age Restrictions'),
            'value_type': 'multiple',
            'options': [
                {'key': '0', 'name': '0', 'default': True},
                {'key': '16', 'name': '16+', 'default': False},
                {'key': '18', 'name': '18+', 'default': False},
            ]
        },
        {
            'key': 'statuses',
            'type': 'select',
            'name': _('Status'),
            'description': _('Filter by Statuses'),
            'value_type': 'multiple',
            'options': [
                {'key': 'ONGOING', 'name': _('Ongoing'), 'default': False},
                {'key': 'FINISHED', 'name': _('Completed'), 'default': False},
                {'key': 'PAUSED', 'name': _('Hiatus'), 'default': False},
            ]
        },
        {
            'key': 'types',
            'type': 'select',
            'name': _('Type'),
            'description': _('Filter by Types'),
            'value_type': 'multiple',
            'options': [
                {'key': 'MANGA_UA', 'name': _('Manga UA'), 'default': False},
                {'key': 'MANGA', 'name': _('Manga'), 'default': False},
                {'key': 'MANHVA', 'name': _('Manhwa'), 'default': False},
                {'key': 'MANHUA', 'name': _('Manhua'), 'default': False},
                {'key': 'WESTERN_COMICS', 'name': _('Western Comic'), 'default': False},
                {'key': 'COMICS', 'name': _('Comic'), 'default': False},
                {'key': 'RANOBE', 'name': _('Light Novel'), 'default': False},
                {'key': 'OTHER', 'name': _('Other'), 'default': False},
            ],
        },
    ]

    headers = {
        'User-Agent': USER_AGENT,
    }

    long_strip_genres = [
        'MANHUA',
        'MANHVA',
    ]

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers = self.headers

    def get_manga_data(self, initial_data):
        """
        Returns manga data from API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.api_manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = resp_data['engName'] if resp_data.get('engName') else resp_data['name']
        if cover_id := resp_data.get('coverImg'):
            data['cover'] = self.image_url.format(cover_id)

        # Details
        if resp_data['status'] == 'FINISHED':
            data['status'] = 'complete'
        elif resp_data['status'] == 'ONGOING':
            data['status'] = 'ongoing'
        elif resp_data['status'] == 'PAUSED':
            data['status'] = 'hiatus'

        if genres := resp_data.get('genres'):
            for genre in genres:
                data['genres'].append(genre['name'])

        if tags := resp_data.get('tags'):
            for tag in tags:
                if not tag['isVerified'] or tag['name'] in data['genres']:
                    continue
                data['genres'].append(tag['name'])

        if category := resp_data.get('category'):
            data['genres'].append(category)

        if writers := resp_data['writers']:
            for writer in writers:
                data['authors'].append(writer['name'])

        if painters := resp_data['painters']:
            for painter in painters:
                if painter['name'] in data['authors']:
                    continue
                data['authors'].append(painter['name'])

        if teams := resp_data.get('teams'):
            for team in teams:
                data['scanlators'].append(team['name'])

        if synopsis := resp_data.get('description'):
            data['synopsis'] = f'{resp_data["name"]}\n\n{synopsis}'

        # Chapters
        r = self.session_get(
            self.api_chapters_url.format(data['slug']),
            headers={
                'Referer': self.manga_url.format(data['slug']),
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        for chapter in reversed(resp_data):
            num_volume = None
            num = None
            name = None

            parts = chapter['name'].split('@#%&;№%#&**#!@')
            size = len(parts)

            if size == 3:
                num_volume, num, name = parts
            elif size == 2:
                num_volume, num = parts
            elif size == 1:
                name = parts[0]

            title = []
            if num_volume:
                title.append(f'Том {num_volume}')
            if num:
                title.append(f'Розділ {num}')
            if name:
                title.append(name)

            ts = chapter['updatedAt'] if chapter.get('updatedAt') else chapter['createdAt']
            data['chapters'].append(dict(
                slug=chapter['id'],
                title=' '.join(title),
                num=num if is_number(num) else None,
                num_volume=num_volume if is_number(num_volume) else None,
                scanlators=[chapter['publisher']['name']] if chapter.get('publisher') else None,
                date=datetime.datetime.fromtimestamp(ts).date(),
            ))

        def sort_func(c):
            if c['num_volume'] or c['num']:
                return '{0:03d}{1:05.1f}'.format(
                    int(c['num_volume']) if c['num_volume'] else 0,
                    float(c['num']) if c['num'] else 0
                )

            return c['slug']

        data['chapters'] = sorted(data['chapters'], key=sort_func)

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data from API

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.api_chapter_url.format(chapter_slug),
            headers={
                'Referer': self.chapter_url.format(manga_slug, chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data = dict(
            pages=[],
        )
        for page in resp_data['pages']:
            data['pages'].append(dict(
                slug=page['content'],
                image=None,
                index=page['order'] + 1,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(self.image_url.format(page['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name='{0:03d}.{1}'.format(page['index'], mime_type.split('/')[-1]),
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, term=None, age_limit=None, statuses=None, types=None, orderby=None):
        def get_page(page):
            params = {
                'limit': 15,
                'offset': (page - 1) * 15,
                'order': 'DESC',
                'ageLimit': ','.join(age_limit) if age_limit else 0,
                'releaseYearFrom': 1980,
                'releaseYearTo': datetime.date.today().year,
            }
            if term:
                params['name'] = term

            if statuses:
                params['status'] = ','.join(statuses)
            if types:
                params['categories'] = ','.join(types)

            if orderby == 'popular':
                params['sortBy'] = 'viewsCount'
            elif orderby == 'latest':
                params['sortBy'] = 'lastChapterCreatedAt'
            else:
                params['sortBy'] = 'createdAt'

            r = self.session_get(
                self.api_search_url,
                params=params,
                headers={
                    'Referer': f'{self.base_url}/',
                }
            )
            if r.status_code != 200:
                return [], False, None

            resp_data = r.json()

            more = page < resp_data['meta']['totalPages'] and page < SEARCH_RESULTS_PAGES

            return resp_data['data'], more, get_response_elapsed(r)

        results = []
        delay = None
        more = True
        page = 1
        while more:
            if delay:
                time.sleep(delay)

            items, more, rtime = get_page(page)
            for item in items:
                results.append(dict(
                    slug=item['id'],
                    name=item['engName'] if item.get('engName') else item['name'],
                    cover=self.image_url.format(item['coverImg']),
                ))

            delay = min(rtime * 4, DOWNLOAD_MAX_DELAY) if rtime else None
            page += 1

        return results

    def get_latest_updates(self, age_limit=None, statuses=None, types=None):
        return self.get_manga_list(age_limit=age_limit, statuses=statuses, types=types, orderby='latest')

    def get_most_populars(self, age_limit=None, statuses=None, types=None):
        return self.get_manga_list(age_limit=age_limit, statuses=statuses, types=types, orderby='popular')

    def search(self, term, age_limit=None, statuses=None, types=None):
        return self.get_manga_list(term=term, age_limit=age_limit, statuses=statuses, types=types)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from collections import OrderedDict
from bs4 import BeautifulSoup
import logging
import requests
import unidecode

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type

SERVER_NAME = 'Central de Mangás'

headers = OrderedDict(
    [
        ('User-Agent', USER_AGENT),
    ]
)

logger = logging.getLogger('komikku.servers.centraldemangas')


class Centraldemangas(Server):
    id = 'centraldemangas'
    name = SERVER_NAME
    lang = 'pt'
    status = 'disabled'

    base_url = 'http://centraldemangas.online'
    search_url = base_url + '/api/titulos'
    manga_url = base_url + '/titulos/{0}'
    chapter_url = base_url + '/titulos/{0}/manga/ler-online-completo/{1}'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers = headers

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        data['name'] = soup.find('h1').text.strip()

        # Details
        container = soup.find('div', class_='relaxed')
        if not container:
            if soup.find('head').find('title').text.startswith('Página não encontrada'):
                logger.warning('Manga info not found: slug may have changed')
            return None

        for element in container.find_all('div', class_='item'):
            label_element = element.find('div', class_='header')
            if not label_element:
                continue

            label = label_element.text.strip()
            value_element = element.find('div', class_='description')

            if label == 'Sinópse':
                cover_img = value_element.img.extract()
                data['cover'] = cover_img.get('src')
                data['synopsis'] = value_element.text.strip()
            elif label in ('Arte', 'Autor'):
                data['authors'].append(value_element.text.strip())
            elif label == 'Gênero':
                for a_element in value_element.find_all('a'):
                    data['genres'].append(a_element.text.strip())
            elif label == 'Scantrad':
                for a_element in value_element.find_all('a'):
                    data['scanlators'].append(a_element.text.strip())
            elif label == 'Status':
                value = value_element.a.text.strip()

                if value == 'Em publicação':
                    data['status'] = 'ongoing'
                elif value == 'Completo':
                    data['status'] = 'complete'
                elif value == 'Cancelado':
                    data['status'] = 'suspended'
                elif value == 'Pausado':
                    data['status'] = 'hiatus'
            elif label == 'Capítulos':
                for tr_element in reversed(value_element.find_all('div', class_='content')[0].table.tbody.find_all('tr')[1:]):
                    class_ = tr_element.get('class')
                    if class_ and class_[0] == 'active':
                        # Skip volumes rows
                        continue
                    tds_elements = tr_element.find_all('td')
                    date = tds_elements[1].small.text.strip()

                    data['chapters'].append(dict(
                        slug=tds_elements[0].a.get('href').split('/')[-1],
                        title=tds_elements[0].a.text.strip(),
                        date=convert_date_string(date, format='%d/%m/%Y') if date != '00/00/0000' else None,
                    ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        # Pages URLs infos are located in the last JS script at the bottom of document
        pages_slugs = None
        pages_url_start = None
        script = soup.find_all('script')[-1].string
        if script:
            for line in script.split('\n'):
                line = line.strip()[:-1]

                if line.startswith('var pages'):
                    # Ex: var pages = ['01','02','03','04','05','06','07','08','09','10','11','12','13','14',]
                    pages_slugs = line.split('=')[1].strip().replace('[', '').replace(']', '').replace("'", '')[:-1].split(',')
                elif line.startswith('var urlSulfix'):
                    # Ex: var urlSulfix = 'http://mangas2016.centraldemangas.com.br/tales_of_demons_and_gods/tales_of_demons_and_gods002-'
                    pages_url_start = line.split('=')[1].strip().replace("'", '')

        if pages_slugs is not None and pages_url_start is not None:
            data = dict(
                pages=[],
            )

            for page_slug in pages_slugs:
                data['pages'].append(dict(
                    slug=None,
                    image='{0}{1}.jpg'.format(pages_url_start, page_slug),
                ))
        else:
            return None

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        # Scrap HTML page to get image url
        url = page['image']

        r = self.session_get(url, headers={'Referer': self.chapter_url.format(manga_slug, chapter_slug)})
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=url.split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_most_populars(self):
        """
        Returns TOP 10 manga
        """
        r = self.session_get(self.base_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.find_all('div', class_='ui red segment')[0].find_all('a')[:-1]:
            results.append(dict(
                name=a_element.text.strip(),
                slug=a_element.get('href').split('/')[-1],
            ))

        return results

    def search(self, term):
        r = self.session_get(self.search_url, headers={'X-Requested-With': 'XMLHttpRequest'})

        if r.status_code == 200:
            data = r.json(strict=False)

            results = []
            term = unidecode.unidecode(term).lower()
            for item in data:
                if term not in unidecode.unidecode(item['title']).lower():
                    continue

                results.append(dict(
                    slug=item['url'].split('/')[-1],
                    name=item['title'],
                ))

            return results

        return None



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import logging

from komikku.models.database import create_db_connection
from komikku.models.database import update_row
from komikku.models.database.mangas import Chapter

logger = logging.getLogger(__name__)


class Download:
    _chapter = None

    STATUSES = dict(
        pending=_('Download pending'),
        downloaded=_('Downloaded'),
        downloading=_('Downloading'),
        canceled=_('Download canceled'),
        error=_('Download error'),
    )

    @classmethod
    def get(cls, id_, db_conn=None):
        if db_conn is None:
            db_conn = create_db_connection()

        row = db_conn.execute('SELECT * FROM downloads WHERE id = ?', (id_,)).fetchone()

        if row is None:
            return None

        d = cls()
        for key in row.keys():
            setattr(d, key, row[key])

        return d

    @classmethod
    def get_by_chapter_id(cls, chapter_id, db_conn=None):
        if db_conn is None:
            db_conn = create_db_connection()

        row = db_conn.execute('SELECT * FROM downloads WHERE chapter_id = ?', (chapter_id,)).fetchone()

        if row:
            d = cls()

            for key in row.keys():
                setattr(d, key, row[key])

            return d

        return None

    @property
    def chapter(self):
        if self._chapter is None:
            self._chapter = Chapter.get(self.chapter_id)

        return self._chapter

    def delete(self, db_conn=None):
        if db_conn is None:
            db_conn = create_db_connection()

        with db_conn:
            db_conn.execute('DELETE FROM downloads WHERE id = ?', (self.id, ))

    def update(self, data, db_conn=None):
        """
        Updates download

        :param data: percent of pages downloaded, errors or status
        :return: True on success False otherwise
        """
        if db_conn is None:
            db_conn = create_db_connection()

        result = False

        with db_conn:
            if update_row(db_conn, 'downloads', self.id, data):
                result = True
                for key in data:
                    setattr(self, key, data[key])

        return result



# SPDX-FileCopyrightText: 2021-2024 Liliana Prikler
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Liliana Prikler <liliana.prikler@gmail.com>

from bs4 import BeautifulSoup
from gettext import gettext as _
import logging
from operator import itemgetter
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.exceptions import NotFoundError
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type

logger = logging.getLogger('komikku.servers.bilibili')

SEARCH_RESULTS_LIMIT = 9

# Dead since early 2024
# A new site, built from a backup, exists https://comicfans.io


class Bilibili(Server):
    id = 'bilibili'
    name = 'Bilibili Comics'
    lang = 'en'
    status = 'disabled'

    base_url = 'https://www.bilibilicomics.com'
    manga_url = base_url + '/detail/mc{}'

    filters = [
        {
            'key': 'price',
            'type': 'select',
            'name': _('Price'),
            'description': _('Filter by price'),
            'value_type': 'single',
            'default': '1',
            'options': [
                {'key': '-1', 'name': _('All')},
                {'key': '1', 'name': _('Free')},
                {'key': '2', 'name': _('Paid')},
            ],
        },
    ]

    def __init__(self):
        query_params = f'?device=pc&platform=web&lang={self.lang}&sys_lang={self.lang}'

        self.api_base_url = self.base_url + '/twirp/comic.v1.Comic'
        self.api_manga_list_url = self.api_base_url + '/ClassPage' + query_params
        self.api_search_url = self.api_base_url + '/Search' + query_params
        self.api_manga_url = self.api_base_url + '/ComicDetail' + query_params
        self.api_chapter_url = self.api_base_url + '/GetImageIndex' + query_params
        self.api_image_token_url = self.api_base_url + '/ImageToken' + query_params

        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({
                'User-Agent': USER_AGENT,
                'Origin': self.base_url,
                'Referer': f'{self.base_url}/',
                'Accept': 'application/json; text/plain; */*',
                'Content-Type': 'application/json;charset=UTF-8',
            })

    def get_manga_data(self, initial_data):
        """
        Returns manga data from API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_post(
            self.api_manga_url,
            json=dict(
                comic_id=int(initial_data['slug']),
            )
        )
        if r.status_code != 200:
            return None

        json_data = r.json()['data']

        data = initial_data.copy()
        data.update(dict(
            name=json_data['title'],
            synopsis=json_data['evaluate'],
            authors=json_data['author_name'],
            scanlators=[],
            genres=json_data['styles'],
            status='complete' if json_data['is_finish'] else 'ongoing',
            cover=json_data['vertical_cover'],
            chapters=[
                dict(
                    slug=str(ep['id']),
                    title='#{} - {}'.format(ep['short_title'], ep['title']),
                    date=convert_date_string(ep['pub_time'].split('T')[0], format='%Y-%m-%d'),
                )
                for ep in sorted(json_data['ep_list'], key=itemgetter('ord'))
                # We don't support user authentication, much less payment,
                # so we can only offer freely available chapters.
                if ep['pay_mode'] == 0
            ],
            server_id=self.id,
        ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data from API

        Currently, only pages are expected.
        """
        r = self.session_post(self.api_chapter_url, json=dict(ep_id=int(chapter_slug)))
        if r.status_code == 404:
            raise NotFoundError
        if r.status_code != 200:
            return None

        images = r.json()['data']['images']

        data = dict(
            pages=[dict(slug=image['path'], image=None) for image in images]
        )

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        r = self.session_post(self.api_image_token_url, json=dict(urls='["{}"]'.format(page['slug'])))
        if r.status_code != 200:
            return None

        data = r.json()['data'][0]

        r = self.session_get(
            data['url'] + '?token=' + data['token'],
            headers={
                'Origin': self.base_url,
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        expected_content_length = int(r.headers['content-length'])
        if len(r.content) != expected_content_length:
            logger.warning(
                'Mismatched content length, expected {0}, got {1}'.format(len(r.content), expected_content_length)
            )
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['slug'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self, price=1):
        return self.search(None, price=price, orderby='latest')

    def get_most_populars(self, price=1):
        return self.search(None, price=price, orderby='populars')

    def is_long_strip(self, _manga_data):
        return True

    def search(self, term, price=1, orderby=None):
        payload = dict(
            area_id=-1,
            is_finish=-1,
            is_free=price,
            page_num=1,
            style_id=-1,
            style_prefer='[]',
        )
        if orderby:
            payload['order'] = 0 if orderby == 'populars' else 1
            payload['page_size'] = SEARCH_RESULTS_LIMIT * 2
        else:
            payload['order'] = -1
            payload['page_size'] = SEARCH_RESULTS_LIMIT
            payload['need_shield_prefer'] = True
            payload['key_word'] = term

        r = self.session_post(
            self.api_manga_list_url if orderby else self.api_search_url,
            json=payload
        )
        if r.status_code != 200:
            return None

        data = r.json()['data']
        if orderby is None:
            data = data['list']

        results = []
        for manga in data:
            if orderby:
                results.append(dict(
                    slug=str(manga['season_id']),
                    name=manga['title'],
                    cover=manga['vertical_cover'],
                ))
            else:
                results.append(dict(
                    slug=str(manga['id']),
                    name=BeautifulSoup(manga['title'], 'lxml').text,
                    cover=manga['vertical_cover'],
                ))

        return results


class Bilibili_es(Bilibili):
    id = 'bilibili_es'
    lang = 'es'


class Bilibili_fr(Bilibili):
    id = 'bilibili_fr'
    lang = 'fr'


class Bilibili_id(Bilibili):
    id = 'bilibili_id'
    lang = 'id'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _


class ServerException(Exception):
    def __init__(self, message):
        self.message = _('Error: {}').format(message)
        super().__init__(self.message)


class ArchiveError(ServerException):
    def __init__(self):
        super().__init__(_('Local archive is corrupt.'))


class ArchiveUnrarMissingError(ServerException):
    def __init__(self):
        super().__init__(_("Unable to extract page. Maybe the 'unrar' tool is missing?"))


class ChallengerError(ServerException):
    def __init__(self):
        super().__init__(_('Failed to complete browser challenge. Please try again.'))


class NotFoundError(ServerException):
    def __init__(self):
        super().__init__(_('No longer exists.'))



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: ISO-morphism <me@iso-morphism.name>

from functools import wraps
import logging

from bs4 import BeautifulSoup
try:
    # This server requires JA3/TLS and HTTP2 fingerprints impersonation
    from curl_cffi import requests
except Exception:
    # Server will be disabled
    requests = None

from komikku.consts import REQUESTS_TIMEOUT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number

logger = logging.getLogger('komikku.servers.mangahub')


def get_api_key(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        server = args[0]
        if not server.api_key:
            server.session_get(server.base_url)
            for name, value in server.session.cookies.items():
                if name == 'mhub_access':
                    server.api_key = value
                    break

        return func(*args, **kwargs)

    return wrapper


class Mangahub(Server):
    id = 'mangahub'
    name = 'MangaHub'
    lang = 'en'
    is_nsfw = True
    status = 'enabled' if requests is not None else 'disabled'

    http_client = 'curl_cffi'

    base_url = 'https://mangahub.io'
    logo_url = base_url + '/favicon-32x32.png'
    api_url = 'https://api.mghcdn.com/graphql'
    manga_url = base_url + '/manga/{0}'
    chapter_url = base_url + '/chapter/{0}/{1}'
    cover_url = 'https://thumb.mghcdn.com/{0}'

    long_strip_genres = ['Webtoon', 'Webtoons', 'LONG STRIP', 'LONG STRIP ROMANCE', ]

    def __init__(self):
        self.api_key = None

        if self.session is None and requests is not None:
            self.session = requests.Session(allow_redirects=True, impersonate='chrome', timeout=(REQUESTS_TIMEOUT, REQUESTS_TIMEOUT * 2))

    @get_api_key
    def get_manga_data(self, initial_data):
        """
        Returns manga data via GraphQL API.

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        query = {
            'query': '{latestPopular(x:m01){id,rank,title,slug,image,latestChapter,isLicensed,updatedDate}manga(x:m01,slug:"%s"){id,rank,title,slug,status,image,latestChapter,author,artist,genres,description,alternativeTitle,mainSlug,isYaoi,isPorn,isSoftPorn,isLicensed,noCoverAd,createdDate,updatedDate,chapters{id,number,title,slug,date}}}' % initial_data['slug']
        }
        r = self.session.post(
            self.api_url,
            json=query,
            headers={
                'Accept': 'application/json',
                'Content-Type': 'application/json',
                'Referer': self.base_url + '/',
                'Origin': self.base_url,
                'x-mhub-access': self.api_key,
            }
        )
        if r.status_code != 200:
            return None

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
            cover=None,
        ))

        manga = r.json()['data']['manga']

        data['name'] = manga['title']
        data['cover'] = self.cover_url.format(manga['image'])

        # Details
        data['authors'] = [author.strip() for author in manga['author'].split(',')]
        for artist in manga['artist'].split(','):
            artist = artist.strip()
            if artist not in data['authors']:
                data['authors'].append(artist)

        data['genres'] = [genre.strip() for genre in manga['genres'].split(',')]

        if manga['status'] == 'ongoing':
            data['status'] = 'ongoing'
        elif manga['status'] == 'completed':
            data['status'] = 'complete'

        data['synopsis'] = manga['description']

        # Chapters
        for chapter in manga['chapters']:
            if chapter['title']:
                title = '#{0} - {1}'.format(chapter['number'], chapter['title'])
            else:
                title = '#{0}'.format(chapter['number'])

            data['chapters'].append(dict(
                slug='chapter-{}'.format(chapter['number']),
                title=title,
                num=chapter['number'] if is_number(chapter['number']) else None,
                date=convert_date_string(chapter['date'].split('T')[0], format='%Y-%m-%d'),
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns chapter's data via GraphQL API

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(manga_slug, chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )
        for element in soup.select('#mangareader img[loading="lazy"]'):
            data['pages'].append(dict(
                slug=None,
                image=element.get('src'),
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Accept': 'image/avif,image/webp,image/png,image/svg+xml,image/*;q=0.8,*/*;q=0.5',
                'Referer': f'{self.base_url}/',
                'Sec-Fetch-Dest': 'image',
                'Sec-Fetch-Mode': 'no-cors',
                'Sec-Fetch-Site': 'cross-site',
            },
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self):
        """
        Returns latest manga
        """
        return self.search('', orderby='latest')

    def get_most_populars(self):
        """
        Returns popular manga
        """
        return self.search('', orderby='popular')

    @get_api_key
    def search(self, term, orderby=None):
        if orderby is not None:
            query = {
                'query': '{search(x:m01,mod:%s,count:true,offset:0){rows{id,rank,title,slug,status,author,genres,image,latestChapter,isLicensed,createdDate},count}}' % orderby.upper()
            }
        else:
            query = {
                'query': '{search(x:m01,q:"%s",limit:10){rows{id,title,slug,image,rank,latestChapter,createdDate}}}' % term
            }

        r = self.session_post(self.api_url, json=query, headers={
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Origin': self.base_url,
            'Referer': self.base_url + '/',
            'x-mhub-access': self.api_key,
        })
        if r.status_code != 200:
            return None

        results = []
        for row in r.json()['data']['search']['rows']:
            results.append(dict(
                slug=row['slug'],
                name=row['title'],
                cover=self.cover_url.format(row['image']),
                last_chapter=row['latestChapter'],
            ))

        return results



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import pytz

from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

from komikku.consts import COVER_WIDTH
from komikku.consts import MISSING_IMG_RESOURCE_PATH
from komikku.consts import TIMEZONE
from komikku.card.synopsis_fading import SynopsisFading
from komikku.models import Category
from komikku.utils import CoverPicture
from komikku.utils import folder_size
from komikku.utils import html_escape


class InfoBox:
    def __init__(self, card):
        self.card = card
        self.window = card.window

        self.title_box = self.card.title_box
        self.cover_box = self.card.cover_box
        self.cover_picture = None
        self.name_label = self.card.name_label
        self.authors_label = self.card.authors_label
        self.status_server_label = self.card.status_server_label
        self.buttons_box = self.card.buttons_box
        self.add_button = self.card.add_button
        self.resume_button = self.card.resume_button
        self.genres_wrapbox = self.card.genres_wrapbox
        self.categories_wrapbox = self.card.categories_wrapbox
        self.scanlators_label = self.card.scanlators_label
        self.chapters_label = self.card.chapters_label
        self.last_update_label = self.card.last_update_label
        self.size_on_disk_label = self.card.size_on_disk_label

        # Synopsis
        self.synopsis_box = self.card.synopsis_box
        self.synopsis_togglebutton = self.card.synopsis_togglebutton
        # To distinguish the `Show Less/More` button from the `Resume` button,
        # we apply `circular` class rather than `pill`
        # This makes it smaller in height, but it requires increasing start and end margins
        togglebutton_label = self.card.synopsis_togglebutton.get_child()
        togglebutton_label.props.margin_start = 24
        togglebutton_label.props.margin_end = 24
        self.synopsis_fading = SynopsisFading()
        self.synopsis_label = Gtk.Label(hexpand=True, xalign=0, wrap=True, wrap_mode=Pango.WrapMode.WORD_CHAR)
        self.synopsis_label.set_css_classes(['document'])
        self.synopsis_fading.set_child(self.synopsis_label)
        self.synopsis_box.prepend(self.synopsis_fading)

        self.add_button.connect('clicked', self.card.on_add_button_clicked)
        self.resume_button.connect('clicked', self.card.on_resume_button_clicked)
        self.synopsis_togglebutton.connect('toggled', self.on_synopsis_togglebutton_toggled)
        self.synopsis_fading.bind_property(
            'faded', self.synopsis_togglebutton, 'visible',
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE
        )

        self.window.breakpoint.add_setter(self.title_box, 'orientation', Gtk.Orientation.VERTICAL)
        self.window.breakpoint.add_setter(self.title_box, 'spacing', 12)
        self.window.breakpoint.add_setter(self.name_label, 'halign', Gtk.Align.CENTER)
        self.window.breakpoint.add_setter(self.name_label, 'justify', Gtk.Justification.CENTER)
        self.window.breakpoint.add_setter(self.status_server_label, 'halign', Gtk.Align.CENTER)
        self.window.breakpoint.add_setter(self.status_server_label, 'justify', Gtk.Justification.CENTER)
        self.window.breakpoint.add_setter(self.authors_label, 'halign', Gtk.Align.CENTER)
        self.window.breakpoint.add_setter(self.authors_label, 'justify', Gtk.Justification.CENTER)
        self.window.breakpoint.add_setter(self.buttons_box, 'orientation', Gtk.Orientation.VERTICAL)
        self.window.breakpoint.add_setter(self.buttons_box, 'spacing', 18)
        self.window.breakpoint.add_setter(self.buttons_box, 'halign', Gtk.Align.CENTER)

    def on_synopsis_togglebutton_toggled(self, button):
        active = button.get_active()
        self.synopsis_togglebutton.set_label(_('Show Less') if active else _('Show More'))
        self.synopsis_fading.set_revealed(active)

    def populate(self):
        manga = self.card.manga

        # Name
        self.name_label.set_text(manga.name)

        # Cover
        if self.cover_picture:
            self.cover_box.remove(self.cover_picture)

        if manga.cover_fs_path is None:
            self.cover_picture = CoverPicture.new_from_resource(MISSING_IMG_RESOURCE_PATH, width=COVER_WIDTH)
            self.card.remove_backdrop()
        else:
            picture = CoverPicture.new_from_file(manga.cover_fs_path, width=COVER_WIDTH)
            if picture:
                self.cover_picture = picture
                self.card.set_backdrop()
            else:
                self.cover_picture = CoverPicture.new_from_resource(MISSING_IMG_RESOURCE_PATH, width=COVER_WIDTH)
                self.card.remove_backdrop()

        self.cover_picture.props.can_shrink = False
        self.cover_picture.add_css_class('cover-dropshadow')
        self.cover_box.append(self.cover_picture)

        # Authors
        authors = html_escape(', '.join(manga.authors)) if manga.authors else _('Unknown author')
        self.authors_label.set_markup(authors)

        # Server (link to server page)
        if not manga.is_local:
            self.status_server_label.set_markup(
                '{0} · <a href="{1}">{2}</a> ({3})'.format(
                    _(manga.STATUSES[manga.status]) if manga.status else _('Unknown status'),
                    manga.server.get_manga_url(manga.slug, manga.url),
                    html_escape(manga.server.name),
                    manga.server.lang.upper() if manga.server.lang else '??'
                )
            )
        else:
            self.status_server_label.set_text('{0} · {1}'.format(_('Unknown status'), _('Local')))

        # Resume button
        if manga.in_library:
            self.add_button.set_visible(False)
            self.resume_button.add_css_class('suggested-action')
        else:
            self.add_button.set_visible(True)
            self.resume_button.remove_css_class('suggested-action')

        # Genres
        if manga.genres:
            self.genres_wrapbox.remove_all()

            for genre in sorted(manga.genres):
                label = Gtk.Label()
                label.set_ellipsize(Pango.EllipsizeMode.END)
                label.set_markup(html_escape(genre))
                label.set_css_classes(['genre-label', 'caption'])
                self.genres_wrapbox.append(label)

            self.genres_wrapbox.get_parent().set_visible(True)
        else:
            self.genres_wrapbox.get_parent().set_visible(False)

        # Categories
        self.set_categories()

        # Scanlators
        if manga.scanlators:
            self.scanlators_label.set_markup(html_escape(', '.join(manga.scanlators)))
            self.scanlators_label.get_parent().set_visible(True)
        else:
            self.scanlators_label.get_parent().set_visible(False)

        # Number of chapters
        self.chapters_label.set_text(str(len(manga.chapters)))

        # Last update date
        if manga.last_update:
            self.last_update_label.set_text(
                manga.last_update.replace(tzinfo=pytz.UTC).astimezone(TIMEZONE).strftime(_('%m/%d/%Y %H:%M'))
            )
            self.last_update_label.get_parent().set_visible(True)
        else:
            self.last_update_label.get_parent().set_visible(False)

        # Disk usage
        self.set_disk_usage()

        # Synopsis
        self.synopsis_fading.set_markup('-')
        if manga.synopsis:
            self.synopsis_box.set_visible(True)
            self.synopsis_togglebutton.set_label(_('Show More'))
            self.synopsis_togglebutton.set_active(False)

            synopsis = manga.synopsis
            if manga.server.donate_url:
                synopsis = f'<a href="{manga.server.donate_url}">{_("Donate")}</a>\n\n{synopsis}'

            self.synopsis_fading.set_markup(synopsis)  # can failed with a warning: parsing markup error
        else:
            self.synopsis_box.set_visible(False)

    def refresh(self):
        self.set_disk_usage()

    def set_categories(self):
        if self.card.manga.categories:
            self.categories_wrapbox.remove_all()

            for category_id in sorted(self.card.manga.categories):
                category = Category.get(category_id)
                label = Gtk.Label()
                label.set_ellipsize(Pango.EllipsizeMode.END)
                label.set_markup(html_escape(category.label))
                label.set_css_classes(['category-label', 'caption'])
                self.categories_wrapbox.append(label)

            self.categories_wrapbox.get_parent().set_visible(True)
        else:
            self.categories_wrapbox.get_parent().set_visible(False)

    def set_disk_usage(self):
        self.size_on_disk_label.set_text(folder_size(self.card.manga.path) or '-')



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.heancms import HeanCMS


class Rezoscans(HeanCMS):
    id = 'rezoscans'
    name = 'Rezo Scans'
    lang = 'en'
    status = 'disabled'  # merge with Qi Scans

    base_url = 'https://rezoscan.org'
    logo_url = 'https://storage.rezoscan.org/upload/2025/06/24/%D8%A7%D9%84%D9%84%D9%88%D8%BA%D9%88-9f96b7a917940f61.webp'
    api_url = 'https://api.rezoscan.org/api'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from copy import deepcopy
from gettext import gettext as _

from gi.repository import Adw
from gi.repository import Gtk

from komikku.models import Settings
from komikku.models.keyring import KeyringHelper
from komikku.servers import LANGUAGES
from komikku.servers.utils import get_server_main_id_by_id
from komikku.servers.utils import get_servers_list
from komikku.utils import html_escape


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/preferences_server_params.ui')
class PreferencesServerParamsSupPage(Adw.NavigationPage):
    __gtype_name__ = 'PreferencesServerParamsSubPage'

    headerbar = Gtk.Template.Child('headerbar')
    page = Gtk.Template.Child('page')

    def __init__(self, data):
        Adw.NavigationPage.__init__(self)

        self.data = data
        self.keyring_helper = KeyringHelper()
        self.settings = Settings.get_default()

        self.headerbar.get_title_widget().set_subtitle(data['name'])

        # Populate
        self.add_credentials_form()
        self.add_params_fields()

    def add_credentials_form(self):
        if not self.data['has_login']:
            return

        group = Adw.PreferencesGroup()
        group.set_title(_('User Account'))
        self.page.add(group)

        if self.data['base_url'] is None:
            # Server has a customizable address/base_url (ex. Komga)
            address_entry = Adw.EntryRow(title=_('Address'))
            address_entry.add_prefix(Gtk.Image.new_from_icon_name('network-server-symbolic'))
            group.add(address_entry)
        else:
            address_entry = None

        username_entry = Adw.EntryRow(title=_('Username'))
        username_entry.add_prefix(Gtk.Image.new_from_icon_name('avatar-default-symbolic'))
        group.add(username_entry)

        password_entry = Adw.PasswordEntryRow(title=_('Password'))
        password_entry.add_prefix(Gtk.Image.new_from_icon_name('dialog-password-symbolic'))
        group.add(password_entry)

        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, margin_top=12, margin_bottom=12, spacing=12)
        group.add(box)

        plaintext_checkbutton = None
        if self.keyring_helper.is_disabled or not self.keyring_helper.has_recommended_backend:
            label = Gtk.Label(hexpand=True)
            label.set_wrap(True)
            if self.keyring_helper.is_disabled:
                label.add_css_class('dimmed')
                label.set_text(_('System keyring service is disabled. Credential cannot be saved.'))
                box.append(label)
            elif not self.keyring_helper.has_recommended_backend:
                if not self.settings.credentials_storage_plaintext_fallback:
                    plaintext_checkbutton = Gtk.CheckButton.new()
                    label.set_text(_('No keyring backends were found to store credential. Use plaintext storage as fallback.'))
                    plaintext_checkbutton.set_child(label)
                    box.append(plaintext_checkbutton)
                else:
                    label.add_css_class('dimmed')
                    label.set_text(_('No keyring backends were found to store credential. Plaintext storage will be used as fallback.'))
                    box.append(label)

        btn = Gtk.Button()
        btn_hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        btn_hbox.set_halign(Gtk.Align.CENTER)
        btn.icon = Gtk.Image(visible=False)
        btn_hbox.append(btn.icon)
        btn_hbox.append(Gtk.Label(label=_('Test')))
        btn.connect(
            'clicked', self.save_credential,
            username_entry, password_entry, address_entry, plaintext_checkbutton
        )
        btn.set_child(btn_hbox)
        box.append(btn)

        credential = self.keyring_helper.get(self.data['main_id'])
        if credential:
            if address_entry is not None:
                address_entry.set_text(credential.address)
            username_entry.set_text(credential.username)
            password_entry.set_text(credential.password)

    def add_params_fields(self):
        defaults = get_server_default_params(self.data)
        if not defaults:
            return

        servers_settings = deepcopy(self.settings.servers_settings)

        if self.data['main_id'] not in servers_settings:
            servers_settings[self.data['main_id']] = {'params': defaults}
        elif 'params' not in servers_settings[self.data['main_id']]:
            servers_settings[self.data['main_id']]['params'] = defaults
        else:
            for key, value in defaults.items():
                if key not in servers_settings[self.data['main_id']]['params']:
                    servers_settings[self.data['main_id']]['params'][key] = value

        def build_select_single(group, param):
            def on_selected(row, _param):
                position = row.get_selected()
                servers_settings[self.data['main_id']]['params'][param['key']] = param['options'][position]['key']
                self.settings.servers_settings = servers_settings

            labels = Gtk.StringList()
            selected_position = 0
            if param['key'] in servers_settings[self.data['main_id']]['params']:
                value = servers_settings[self.data['main_id']]['params'][param['key']]
            else:
                value = param['default']
            for index, option in enumerate(param['options']):
                labels.append(option['name'])
                if option['key'] == value:
                    selected_position = index

            row = Adw.ComboRow(title=param['name'], subtitle=param['description'])
            row.set_use_markup(True)
            row.set_model(labels)
            row.set_selected(selected_position)
            row.connect('notify::selected', on_selected)

            group.add(row)

            return group

        def build_select_multiple(group, param):
            def on_active(row, _param, key):
                if row.get_active():
                    servers_settings[self.data['main_id']]['params'][param['key']].append(key)
                else:
                    servers_settings[self.data['main_id']]['params'][param['key']].remove(key)
                self.settings.servers_settings = servers_settings

            group.set_title(param['name'])
            group.set_description(param['description'])

            for option in param['options']:
                row = Adw.SwitchRow(title=option['name'])
                row.set_use_markup(True)
                row.set_active(option['key'] in servers_settings[self.data['main_id']]['params'].get(param['key'], defaults[param['key']]))
                row.connect('notify::active', on_active, option['key'])

                group.add(row)

            return group

        def build_switch(group, param):
            def on_active(row, _param, key):
                servers_settings[self.data['main_id']]['params'][param['key']] = row.get_active()
                self.settings.servers_settings = servers_settings

            row = Adw.SwitchRow(title=param['name'])
            row.set_use_markup(True)
            row.set_subtitle(param['description'])
            if param['key'] in servers_settings[self.data['main_id']]['params']:
                row.set_active(servers_settings[self.data['main_id']]['params'][param['key']])
            else:
                row.set_active(param['default'])
            row.connect('notify::active', on_active, param['key'])

            group.add(row)

            return group

        for param in self.data['params']:
            group = Adw.PreferencesGroup()

            if param['type'] == 'select':
                if param['value_type'] == 'single':
                    build_select_single(group, param)
                elif param['value_type'] == 'multiple':
                    build_select_multiple(group, param)
                else:
                    raise ValueError('Invalid select value_type')  # noqa: TC003
            elif param['type'] == 'checkbox':
                build_switch(group, param)
            else:
                raise ValueError('Invalid param type')  # noqa: TC003

            self.page.add(group)

        # Add some filters (include_in_settings key == True)
        # Allow to define default values for server filters in Explorer (search)
        # and to configure the server's behavior
        for param in self.data['filters']:
            if not param.get('include_in_settings'):
                continue

            group = Adw.PreferencesGroup()

            if param['type'] == 'select':
                if param['value_type'] == 'single':
                    build_select_single(group, param)
                elif param['value_type'] == 'multiple':
                    build_select_multiple(group, param)
                else:
                    raise ValueError('Invalid select value_type')  # noqa: TC003
            elif param['type'] == 'checkbox':
                build_switch(group, param)
            else:
                raise ValueError('Invalid param type')  # noqa: TC003

            self.page.add(group)

    def save_credential(self, button, username_entry, password_entry, address_entry, plaintext_checkbutton):
        class_ = getattr(self.data['module'], self.data['main_id'].capitalize())

        username = username_entry.get_text()
        password = password_entry.get_text()
        if address_entry is not None:
            address = address_entry.get_text().strip()
            if not address.startswith(('http://', 'https://')):
                return

            server = class_(username=username, password=password, address=address)
        else:
            address = None
            server = class_(username=username, password=password)

        button.icon.set_visible(True)
        if server.logged_in:
            button.icon.set_from_icon_name('checkmark-symbolic')
            if self.keyring_helper.is_disabled or plaintext_checkbutton is not None and not plaintext_checkbutton.get_active():
                return

            if plaintext_checkbutton is not None and plaintext_checkbutton.get_active():
                # Save user agrees to save credentials in plaintext
                self.parent.credentials_storage_plaintext_fallback_switch.set_active(True)

            self.keyring_helper.store(self.data['main_id'], username, password, address)
        else:
            button.icon.set_from_icon_name('computer-fail-symbolic')


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/preferences_servers_languages.ui')
class PreferencesServersLanguagesSubPage(Adw.NavigationPage):
    __gtype_name__ = 'PreferencesServersLanguagesSubPage'

    group = Gtk.Template.Child('group')

    def __init__(self, parent):
        Adw.NavigationPage.__init__(self)

        self.parent = parent
        self.window = self.parent.window
        self.settings = Settings.get_default()

    def clear(self):
        row = self.group.get_row(0)
        while row:
            next_row = row.get_next_sibling()
            self.group.remove(row)
            row = next_row

    def on_language_activated(self, switchrow, _gparam, code):
        if switchrow.get_active():
            self.settings.add_servers_language(code)
        else:
            self.settings.remove_servers_language(code)

        # Update Explorer servers page
        if self.window.explorer.servers_page in self.window.navigationview.get_navigation_stack():
            self.window.explorer.servers_page.populate()

    def populate(self, *args):
        self.clear()

        servers_languages = self.settings.servers_languages

        for code, language in LANGUAGES.items():
            switchrow = Adw.SwitchRow()
            switchrow.set_title(language)
            switchrow.set_active(code in servers_languages)
            switchrow.connect('notify::active', self.on_language_activated, code)

            self.group.add(switchrow)

        self.parent.push_subpage(self)


@Gtk.Template.from_resource('/info/febvre/Komikku/ui/preferences_servers_settings.ui')
class PreferencesServersSettingsSubPage(Adw.NavigationPage):
    __gtype_name__ = 'PreferencesServersSettingsSubPage'

    group = Gtk.Template.Child('group')

    def __init__(self, parent):
        Adw.NavigationPage.__init__(self)

        self.parent = parent
        self.window = self.parent.window
        self.settings = Settings.get_default()

    def clear(self):
        row = self.group.get_row(0)
        while row:
            next_row = row.get_next_sibling()
            self.group.remove(row)
            row = next_row

    def on_server_activated(self, row, _gparam, server_main_id):
        if isinstance(row, Adw.ExpanderRow):
            self.settings.toggle_server(server_main_id, row.get_enable_expansion())
        else:
            self.settings.toggle_server(server_main_id, row.get_active())

        # Update explorer servers page
        if self.window.explorer.servers_page in self.window.navigationview.get_navigation_stack():
            self.window.explorer.servers_page.populate()

    def on_server_language_activated(self, switch_button, _gparam, server_main_id, lang):
        self.settings.toggle_server_lang(server_main_id, lang, switch_button.get_active())

        # Update explorer servers page
        if self.window.explorer.servers_page in self.window.navigationview.get_navigation_stack():
            self.window.explorer.servers_page.populate()

    def populate(self, *args):
        settings = self.settings.servers_settings
        languages = self.settings.servers_languages

        self.clear()

        servers = get_servers_list(order_by=('name', 'lang'))
        self.window.application.logger.info('{0} servers found'.format(len(servers)))

        servers_data = {}
        for server_data in servers:
            main_id = get_server_main_id_by_id(server_data['id'])

            if main_id not in servers_data:
                servers_data[main_id] = dict(
                    main_id=main_id,
                    name=server_data['name'],
                    description=server_data['description'],
                    module=server_data['module'],
                    base_url=server_data['base_url'],
                    has_login=server_data['has_login'],
                    is_nsfw=server_data['is_nsfw'],
                    is_nsfw_only=server_data['is_nsfw_only'],
                    filters=server_data['filters'],
                    params=server_data['params'],
                    langs=[],
                    langs_enabled=[],
                )

            if server_data['lang']:
                servers_data[main_id]['langs'].append(server_data['lang'])
                if not languages or server_data['lang'] in languages:
                    servers_data[main_id]['langs_enabled'].append(server_data['lang'])

        for server_main_id, server_data in servers_data.items():
            if server_data['langs'] and not server_data['langs_enabled']:
                continue

            server_settings = settings.get(server_main_id)

            server_allowed = not server_data['is_nsfw'] or (server_data['is_nsfw'] and self.settings.nsfw_content)
            server_allowed &= not server_data['is_nsfw_only'] or (server_data['is_nsfw_only'] and self.settings.nsfw_only_content)
            server_enabled = server_settings is None or server_settings.get('enabled', True)

            if len(server_data['langs_enabled']) > 1:
                vbox = Gtk.Box(
                    orientation=Gtk.Orientation.VERTICAL,
                    margin_start=12, margin_top=6, margin_end=12, margin_bottom=6,
                    spacing=12
                )

                row = Adw.ExpanderRow()
                row.set_title(html_escape(server_data['name']))
                if server_data['is_nsfw'] or server_data['is_nsfw_only']:
                    row.set_subtitle(_('18+'))
                row.set_enable_expansion(server_enabled)
                row.set_sensitive(server_allowed)
                row.connect('notify::enable-expansion', self.on_server_activated, server_main_id)
                row.add_row(vbox)

                for lang in server_data['langs_enabled']:
                    lang_enabled = server_settings is None or server_settings.get('langs', {}).get(lang, True)

                    hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, margin_top=6, margin_bottom=6, spacing=12)

                    label = Gtk.Label(label=LANGUAGES[lang], xalign=0, hexpand=True)
                    hbox.append(label)

                    switch = Gtk.Switch.new()
                    switch.set_active(lang_enabled)
                    switch.connect('notify::active', self.on_server_language_activated, server_main_id, lang)
                    hbox.append(switch)

                    vbox.append(hbox)

                if server_data['params'] or server_data['has_login']:
                    params_btn = Gtk.Button(icon_name='server-settings-symbolic', valign=Gtk.Align.CENTER)
                    params_btn.add_css_class('circular')
                    params_btn.props.margin_end = 33
                    params_btn.connect('clicked', self.push_server_params_subpage, server_data)
                    row.add_suffix(params_btn)

                self.group.add(row)
            else:
                if not server_data['langs_enabled']:
                    # Server has no languages (Local for ex.)
                    # Display it only if it has login or parameters
                    if not server_data['params'] and not server_data['has_login']:
                        continue

                    lang = None
                    lang_enabled = True
                else:
                    lang = server_data['langs_enabled'][0]
                    lang_enabled = server_settings is None or server_settings.get('langs', {}).get(lang, True)

                row = Adw.ActionRow()
                row.set_sensitive(server_allowed)
                row.set_title(html_escape(server_data['name']))
                if lang:
                    subtitle = [LANGUAGES[lang]]
                elif server_data['description']:
                    subtitle = [server_data['description']]
                else:
                    subtitle = []
                if server_data['is_nsfw'] or server_data['is_nsfw_only']:
                    subtitle.append(_('18+'))
                if subtitle:
                    row.set_subtitle(' · '.join(subtitle))

                if server_data['params'] or server_data['has_login']:
                    params_btn = Gtk.Button(icon_name='server-settings-symbolic', valign=Gtk.Align.CENTER)
                    params_btn.add_css_class('circular')
                    params_btn.props.margin_end = 6
                    params_btn.connect('clicked', self.push_server_params_subpage, server_data)
                    row.add_suffix(params_btn)

                switch = Gtk.Switch(valign=Gtk.Align.CENTER)
                switch.set_active(server_enabled and server_allowed and lang_enabled)
                row.add_suffix(switch)
                row.set_activatable_widget(switch)

                if len(server_data['langs']) > 1:
                    switch.connect('notify::active', self.on_server_language_activated, server_main_id, lang)
                else:
                    switch.connect('notify::active', self.on_server_activated, server_main_id)

                self.group.add(row)

        self.parent.push_subpage(self)

    def push_server_params_subpage(self, _btn, data):
        self.parent.push_subpage(PreferencesServerParamsSupPage(data))


def get_server_default_params(data):
    params = {}

    if not data['params'] and not data['filters']:
        return params

    for param in data['params']:
        if param['type'] == 'select' and param['value_type'] == 'multiple':
            params[param['key']] = [option['key'] for option in param['options'] if option['default']]
        else:
            params[param['key']] = param['default']

    # Add default values of filters that are included in server settings
    for param in data['filters']:
        if not param.get('include_in_settings'):
            continue

        if param['type'] == 'select' and param['value_type'] == 'multiple':
            params[param['key']] = [option['key'] for option in param['options'] if option['default']]
        else:
            params[param['key']] = param['default']

    return params



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import datetime
from gettext import gettext as _
from io import BytesIO
import logging
import os
import rarfile
import tarfile
import xml.etree.ElementTree as ET
import zipfile

from ebooklib import epub
from ebooklib import ITEM_COVER
from ebooklib import ITEM_IMAGE
from natsort import natsorted as sorted
from pypdf import PdfReader

from komikku.servers import Server
from komikku.servers.exceptions import ArchiveError
from komikku.servers.exceptions import ArchiveUnrarMissingError
from komikku.servers.exceptions import ServerException
from komikku.servers.utils import convert_date_string
from komikku.utils import concat_images_vertically
from komikku.utils import get_buffer_mime_type
from komikku.utils import get_data_dir
from komikku.utils import get_file_mime_type

IMG_EXTENSIONS = ['bmp', 'gif', 'jpg', 'jpeg', 'jxl', 'png', 'tif', 'tiff', 'webp']

logger = logging.getLogger('komikku.servers.local')


def is_archive(path):
    if zipfile.is_zipfile(path):  # CBZ, EPUB
        return True
    if rarfile.is_rarfile(path):
        return True
    if tarfile.is_tarfile(path):
        return True
    if get_file_mime_type(path) == 'application/pdf':
        return True
    return False


class Archive:
    def __init__(self, path):
        try:
            if get_file_mime_type(path) == 'application/epub+zip':
                self.obj = EPUB(path)

            elif zipfile.is_zipfile(path):
                self.obj = CBZ(path)

            elif rarfile.is_rarfile(path):
                self.obj = CBR(path)

            elif tarfile.is_tarfile(path):
                self.obj = CBT(path)

            elif get_file_mime_type(path) == 'application/pdf':
                self.obj = PDF(path)

        except Exception as e:
            logger.exception(f'Bad/corrupt archive: {path}')
            raise ArchiveError from e

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.obj.close()

    def get_info(self):
        if hasattr(self.obj, 'get_info'):
            return self.obj.get_info()

        # Parse ComicInfo.xml if exists
        data = dict(
            title=None,
            volume=None,
            number=None,
            authors=set(),
            translators=set(),
            genres=set(),
            synopsis=None,
            day=None,
            month=None,
            year=None,
        )
        info_xml = self.get_name_buffer('ComicInfo.xml')
        if not info_xml:
            return data

        tree = ET.ElementTree(ET.fromstring(info_xml))
        root = tree.getroot()

        for info in root.findall('./'):
            if not info.text:
                continue

            if info.tag == 'Title':
                data['title'] = info.text
            elif info.tag == 'Volume':
                data['volume'] = info.text.strip()
            elif info.tag == 'Number':
                data['number'] = info.text.strip()
            elif info.tag in ('Colorist', 'CoverArtist', 'Inker', 'Letterer', 'Penciller', 'Writer'):
                for author in info.text.split(','):
                    data['authors'].add(author.strip())
            elif info.tag == 'Translator':
                for translator in info.text.split(','):
                    data['translators'].add(translator.strip())
            elif info.tag == 'Genre':
                for genre in info.text.split(','):
                    data['genres'].add(genre.strip())
            elif info.tag == 'Manga' and info.text.strip() == 'Yes':
                data['genres'].add('Manga')
            elif info.tag == 'Summary':
                data['synopsis'] = info.text.strip()
            elif info.tag in ('Day', 'Month', 'Year'):
                data[info.tag.lower()] = int(info.text.strip())

        return data

    def get_namelist(self):
        names = []
        for name in self.obj.get_namelist():
            _root, ext = os.path.splitext(name)
            if ext[1:].lower() in IMG_EXTENSIONS:
                names.append(name)

        return sorted(names)

    def get_name_buffer(self, name):
        return self.obj.get_name_buffer(name)


class CBR:
    """Comic Book Rar (CBR) format"""

    def __init__(self, path):
        self.path = path
        self.archive = rarfile.RarFile(self.path)

    def close(self):
        self.archive.close()

    def get_namelist(self):
        return self.archive.namelist()

    def get_name_buffer(self, name):
        try:
            return self.archive.read(name)
        except rarfile.NoRarEntry as e:
            logger.info(f'{self.path}: {e}')
            return None
        except (rarfile.BadRarFile, rarfile.RarCannotExec) as e:
            # Strangely, we are getting BadRarFile exceptions when unrar is not installed
            logger.exception('Failed to execute unrar command')
            raise ArchiveUnrarMissingError from e
        except Exception as e:
            logger.info(f'{self.path}: {e}')
            raise ServerException(e) from e


class CBT:
    """Comic Book Tar (CBT) format"""

    def __init__(self, path):
        self.path = path
        self.archive = tarfile.open(self.path)

    def close(self):
        self.archive.close()

    def get_namelist(self):
        return self.archive.getnames()

    def get_name_buffer(self, name):
        try:
            with self.archive.extractfile(name) as fp:
                buf = fp.read()
            return buf
        except KeyError as e:
            logger.info(f'{self.path}: {e}')
            return None
        except Exception as e:
            logger.info(f'{self.path}: {e}')
            raise ServerException(e) from e


class CBZ:
    """Comic Book Zip (CBZ) format

    Can also handle EPUB archives if images are well named (use of zero as a prefix to numbers to keep correct order)
    """

    def __init__(self, path):
        self.path = path
        self.archive = zipfile.ZipFile(self.path, allowZip64=True)

    def close(self):
        self.archive.close()

    def get_namelist(self):
        return self.archive.namelist()

    def get_name_buffer(self, name):
        try:
            return self.archive.read(name)
        except KeyError as e:
            logger.info(f'{self.path}: {e}')
            return None
        except Exception as e:
            logger.info(f'{self.path}: {e}')
            raise ServerException(e) from e


class EPUB:
    """EPUB format"""

    def __init__(self, path):
        self.path = path

        reader = epub.EpubReader(path, None)
        self.archive = reader.load()

    def close(self):
        return

    def get_info(self):
        data = dict(
            title=self.archive.get_metadata('DC', 'title')[0][0],
            volume=None,
            number=None,
            authors=set(),
            translators=set(),
            genres=set(),
            synopsis=None,
            day=None,
            month=None,
            year=None,
        )

        if creators := self.archive.get_metadata('DC', 'creator'):
            for creator in creators:
                data['authors'].add(creator[0])

        if descriptions := self.archive.get_metadata('DC', 'description'):
            data['synopsis'] = []
            for description in descriptions:
                for line in description[0].split('\n'):
                    data['synopsis'].append(line.strip())

            data['synopsis'] = '\n'.join(data['synopsis'])

        if date := self.archive.get_metadata('DC', 'date'):
            if len(date[0][0]) == 8:
                data['year'] = int(date[0][0][:4])
                data['month'] = int(date[0][0][4:6])
                data['day'] = int(date[0][0][6:8])
            elif date := convert_date_string(date[0][0]):
                data['year'] = date.year
                data['month'] = date.month
                data['day'] = date.day

        if position := self.archive.get_metadata('OPF', 'group-position'):
            if '.' in position[0][1]['content']:
                data['volume'], data['number'] = position[0][1]['content'].split('.')
            else:
                data['volume'] = position[0][1]['content']

        return data

    def get_namelist(self):
        namelist = []

        for image in self.archive.get_items_of_type(ITEM_COVER):
            namelist.append(image.get_name())

        for image in self.archive.get_items_of_type(ITEM_IMAGE):
            namelist.append(image.get_name())

        return namelist

    def get_name_buffer(self, name):
        try:
            return self.archive.get_item_with_href(name).get_content()
        except Exception as e:
            logger.info(f'{self.path}: {e}')
            raise ServerException(e) from e


class PDF:
    """PDF format"""

    def __init__(self, path):
        self.path = path
        self.archive = PdfReader(self.path)

    def close(self):
        self.archive.close()

    def get_namelist(self):
        names = []
        for page_index, page in enumerate(self.archive.pages):
            if page.images:
                # Page image can be split vertically into several images
                images_keys = sorted(page.images.keys())
                name = page.images[images_keys[0]].name
                names.append(f'{page_index}:{",".join(images_keys)}:{name}')  # noqa: E231

        return names

    def get_name_buffer(self, name):
        try:
            page_index, images_keys, _name = name.split(':')
            page_index = int(page_index)
            images_keys = images_keys.split(',')

        except ValueError as e:
            logger.info(f'{self.path}: {e}')
            return None

        except Exception as e:
            logger.info(f'{self.path}: {e}')
            raise ServerException(e) from e

        page = self.archive.get_page(page_index)

        if len(images_keys) == 1:
            buffer = page.images[images_keys[0]].data
        else:
            image = concat_images_vertically(*[page.images[key].image for key in images_keys])

            with BytesIO() as io_buffer:
                image.save(io_buffer, 'JPEG')
                buffer = io_buffer.getvalue()

            image.close()

        return buffer


class Local(Server):
    id = 'local'
    name = 'Local'
    description = _('Comics stored locally as archives in CBZ/CBR/CBT or PDF formats')
    lang = ''

    def get_image(self, data, etag=None):
        """ Returns cover image """
        if data is None:
            return None, None, None

        with Archive(data['path']) as archive:
            buffer = archive.get_name_buffer(data['name'])
        if buffer is None:
            return None, None, None

        mime_type = get_buffer_mime_type(buffer)
        if not mime_type.startswith('image'):
            return None, None, None

        return buffer, None, None

    def get_manga_data(self, initial_data):
        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            chapters=[],
            synopsis=None,
            cover=None,
            server_id=self.id,
            url=None,
        ))

        dir_path = os.path.join(get_data_dir(), self.id, data['slug'])
        if not os.path.exists(dir_path):
            return None

        # Chapters
        for _path, _dirs, files in os.walk(dir_path):
            for file in sorted(files):
                path = os.path.join(dir_path, file)
                if not is_archive(os.path.join(dir_path, file)):
                    continue

                try:
                    with Archive(path) as archive:
                        names = archive.get_namelist()

                        # Used some chapters/volumes info to populate comic data
                        info = archive.get_info()
                        for genre in info['genres']:
                            if genre not in data['genres']:
                                data['genres'].append(genre)
                        for author in info['authors']:
                            if author not in data['authors']:
                                data['authors'].append(author)
                        for translator in info['translators']:
                            if translator not in data['scanlators']:
                                data['scanlators'].append(translator)
                        if not data['synopsis'] and info['synopsis']:
                            data['synopsis'] = info['synopsis']

                        # Cover is by default 1st page of 1st chapter/volume (archive)
                        if data['cover'] is None:
                            data['cover'] = dict(
                                path=path,
                                name=names[0],
                            )

                        title = info['title'] or os.path.splitext(file)[0]
                        if info['number']:
                            title = f'{info["number"]} - {title}'
                        if info['volume']:
                            title = f'{title} ({info["volume"]})'

                        date = datetime.date(info['year'], info['month'] or 1, info['day'] or 1) if info['year'] else None

                        chapter = dict(
                            slug=file,
                            title=title,
                            date=date,
                            scanlators=list(info['translators']),
                        )

                        data['chapters'].append(chapter)
                except Exception:
                    logger.exception(f'Failed to retrieve chapters of {data["name"]}')

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        path = os.path.join(get_data_dir(), self.id, manga_slug, chapter_slug)
        if not os.path.exists(path):
            return None

        with Archive(path) as archive:
            names = archive.get_namelist()

        data = dict(
            pages=[],
        )
        for name in names:
            data['pages'].append(dict(
                slug=name,
                image=None,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        path = os.path.join(get_data_dir(), self.id, manga_slug, chapter_slug)
        if not os.path.exists(path):
            return None

        with Archive(path) as archive:
            content = archive.get_name_buffer(page['slug'])

        mime_type = get_buffer_mime_type(content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=content,
            mime_type=mime_type,
            name=page['slug'],
        )

    def get_manga_url(self, slug, url):
        return None

    def get_latest_updates(self):
        """
        Returns list of latest updated manga
        """
        dir_path = os.path.join(get_data_dir(), self.id)

        result = {}
        for name in os.listdir(dir_path):
            manga_folder = os.path.join(dir_path, name)

            if not os.path.isdir(manga_folder):
                continue

            result[os.path.getmtime(manga_folder)] = dict(
                slug=name,
                name=name,
            )

        return [item for key, item in sorted(result.items(), reverse=True)][:100]

    def search(self, term):
        dir_path = os.path.join(get_data_dir(), self.id)

        result = []
        for name in sorted(os.listdir(dir_path)):
            if not os.path.isdir(os.path.join(dir_path, name)):
                continue

            if term and term.lower() not in name.lower():
                continue

            result.append(dict(
                slug=name,
                name=name,
            ))

        return result



# SPDX-FileCopyrightText: 2023-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.utils import get_buffer_mime_type


class Animesama(Server):
    id = 'animesama'
    name = 'Anime-Sama'
    lang = 'fr'

    base_url = 'https://anime-sama.to'
    logo_url = base_url + '/img/favicon.ico?v=4'
    search_url = base_url + '/catalogue/'
    manga_url = base_url + '/catalogue/{0}/'
    chapter_url = base_url + '/catalogue/{0}/scan/vf/'
    api_chapters_url = base_url + '/s2/scans/get_nb_chap_et_img.php'
    image_url = base_url + '/s2/scans/{0}/{1}/{2}.jpg'
    cover_url = 'https://raw.githubusercontent.com/Anime-Sama/IMG/img/contenu/{0}.jpg'

    long_strip_genres = ['Webcomic']

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data from API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],  # not available
            genres=[],
            status=None,
            cover=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        data['name'] = soup.select_one('.oeuvre-right h1').text.strip()
        data['cover'] = self.cover_url.format(data['slug'])

        if element := soup.select_one('.info-grid > .info-lbl:-soup-contains("Créateur") ~ span'):
            data['authors'].append(element.text.strip())

        if element := soup.select_one('.info-grid > .info-lbl:-soup-contains("État") ~ span'):
            status = element.text.strip()
            if status == 'En cours':
                data['status'] = 'ongoing'
            elif status == 'Terminé':
                data['status'] = 'complete'

        for element in soup.select('.genre-pill'):
            data['genres'].append(element.text.strip())

        if element := soup.select_one('div h2:-soup-contains("Synopsis") ~ p'):
            data['synopsis'] = element.text.strip()

        # Chapters
        oeuvre = self.get_manga_oeuvre(data['slug'])
        if oeuvre is None:
            return None

        r = self.session_get(
            self.api_chapters_url,
            params={
                'oeuvre': oeuvre,
            },
            headers={
                'Referer': self.manga_url.format(data['slug']),
            }
        )
        if r.status_code != 200:
            return None

        rjson = r.json()
        if 'error' in rjson:
            return None

        for num in rjson:
            data['chapters'].append(dict(
                slug=num,
                title=f'Chapitre {num}',
                num=num,
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        oeuvre = self.get_manga_oeuvre(manga_slug)
        if oeuvre is None:
            return None

        # Get chapter images
        r = self.session_get(
            self.api_chapters_url,
            params={
                'oeuvre': oeuvre,
            },
            headers={
                'Referer': self.chapter_url.format(manga_slug),
            }
        )
        if r.status_code != 200:
            return None

        rjson = r.json()
        if 'error' in rjson:
            return None

        data = dict(
            pages=[],
        )
        for index in range(1, rjson[chapter_slug] + 1):
            data['pages'].append(dict(
                slug=None,
                image=self.image_url.format(oeuvre, chapter_slug, index),
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': self.chapter_url.format(manga_slug),
            },
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_oeuvre(self, slug):
        r = self.session_get(self.chapter_url.format(slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        if element := soup.select_one('#titreOeuvre'):
            return element.text  # beware, no strip here

        return None

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self, types=None, statuses=None):
        r = self.session_get(self.base_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('#containerAjoutsScans > div'):
            a_element = element.a
            img_element = a_element.img

            if last_element := element.select_one('.card-info .info-text'):
                last_chapter = last_element.text.strip()
            else:
                last_chapter = None

            results.append(dict(
                slug=a_element.get('href').split('/')[-4],
                name=a_element.select_one('.card-title').text.strip(),
                cover=img_element.get('src'),
                last_chapter=last_chapter,
            ))

        return results

    def search(self, term):
        params = {
            'search': term,
            'type[]': 'Scans',
        }

        r = self.session_get(
            self.search_url,
            params=params,
            headers={
                'Referer': self.search_url,
            }
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for element in soup.select('#list_catalog > div'):
            a_element = element.a
            img_element = a_element.img

            url = a_element.get('href').strip()
            if url[-1] == '/':
                slug = url.split('/')[-2]
            else:
                slug = url.split('/')[-1]

            results.append(dict(
                slug=slug,
                name=a_element.select_one('.card-title').text.strip(),
                cover=img_element.get('src'),
            ))

        return results



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import is_number


class Shadowmanga(Server):
    id = 'shadowmanga'
    name = 'Shadow Manga'
    lang = 'es'
    is_nsfw = True

    base_url = 'https://shademanga.com'
    logo_url = base_url + '/favicon.png'

    manga_url = base_url + '/serie/local/{0}'
    chapter_url = base_url + '/reader/local/{0}'

    api_url = base_url + '/api/series-locales'
    api_latest_updates_url = api_url + '/novedades'
    api_most_popular = api_url + '/popular'
    api_search_url = api_url + '/search-candidates'
    api_manga_url = api_url + '/{0}'
    api_chapter_url = api_url + '/{0}/capitulos/{1}/paginas'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns serie data using API

        Initial data should contain at least serie's slug (provided by search)
        """
        r = self.session_get(
            self.api_manga_url.format(initial_data['slug']),
            headers={
                'Referer': self.manga_url.format(initial_data['slug']),
            }
        )
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data = initial_data.copy()
        data.update({
            'name': resp_data['titulo'],
            'authors': [],
            'scanlators': [],
            'genres': [],
            'status': None,
            'synopsis': resp_data['descripcion'],
            'chapters': [],
            'server_id': self.id,
            'cover': resp_data['portadaUrl'],
        })

        # Authors
        if author := resp_data.get('autor'):
            data['authors'].append(author)

        # Genres
        if genres := resp_data.get('generos'):
            for genre in genres.strip().split(','):
                genre = genre.strip()
                if genre:
                    data['genres'].append(genre)

        # Status
        status = resp_data['estado']
        if status == 'En curso':
            data['status'] = 'ongoing'
        elif status == 'Completado':
            data['status'] = 'complete'
        elif status == 'Pausada':
            data['status'] = 'hiatus'

        # Chapters
        chapters = []
        for chapter in sorted(resp_data['capitulos'], key=lambda i: i['orden']):
            num = chapter['numeroCapitulo']
            title = f'Cap. {num}'
            if chapter['titulo'] and chapter['titulo'] != title:
                title = f'{title} - {chapter["titulo"]}'
            chapters.append({
                'slug': chapter['id'],
                'title': title,
                'num': num if is_number(num) else None,
                'date': convert_date_string(chapter['fechaSubida'].split('T')[0], format='%Y-%m-%d'),
            })

        data['chapters'] = chapters

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns serie chapter data using API

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.api_chapter_url.format(manga_slug, chapter_slug),
            headers={
                'Referer': self.chapter_url.format(chapter_slug),
            }
        )
        if r.status_code != 200:
            return None

        data = {
            'pages': [],
        }
        for url in r.json()['paginas']:
            data['pages'].append({
                'slug': None,
                'image': url,
            })

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': page['image'].split('/')[-1],
        }

    def get_manga_url(self, slug, url):
        """
        Returns serie absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self):
        r = self.session_get(
            self.api_latest_updates_url,
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        result = {}
        for genre in r.json():
            for serie in genre['series']:
                if serie['id'] in result:
                    continue
                result[serie['id']] = {
                    'name': serie['titulo'],
                    'slug': serie['id'],
                    'cover': serie['portadaUrl'],
                }

        return list(result.values())

    def get_most_populars(self):
        r = self.session_get(
            self.api_most_popular,
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        result = {}
        for genre in r.json():
            for serie in genre['series']:
                if serie['id'] in result:
                    continue
                result[serie['id']] = {
                    'name': serie['titulo'],
                    'slug': serie['id'],
                    'cover': serie['portadaUrl'],
                }

        return list(result.values())

    def search(self, term):
        r = self.session_get(
            self.api_search_url,
            params={
                'q': term,
                'includeAdult': 'true',
                'showSinPortada': 'false',
                'take': 120,
            },
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        result = []
        for serie in r.json():
            result.append({
                'name': serie['titulo'],
                'slug': serie['id'],
                'cover': serie['portadaUrl'],
            })

        return result



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# Supported servers:
# FMTeam [FR]
# GTO The Great Site [IT]
# Lupi Team [IT]
# Phoenix Scans [IT]
# Tutto Anime Manga [IT]

import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type


class PizzaReader(Server):
    base_url: str

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data from API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.base_url + '/api/comics/' + initial_data['slug'])
        if r.status_code != 200:
            return None

        resp_data = r.json()['comic']

        data = initial_data.copy()
        data.update(dict(
            name=resp_data['title'],
            cover=resp_data['thumbnail_small'],
            authors=[],
            scanlators=[],  # not available
            genres=[genre['name'] for genre in resp_data['genres']],
            status=None,
            synopsis=resp_data['description'],
            chapters=[],
            server_id=self.id,
        ))

        if resp_data['author']:
            data['authors'].append(resp_data['author'])
        if resp_data['artist'] and resp_data['artist'] not in data['authors']:
            data['authors'].append(resp_data['artist'])

        if resp_data['status'].lower().startswith(('in corso', 'on going')):
            data['status'] = 'ongoing'
        else:
            data['status'] = 'complete'

        # Chapters
        for chapter in reversed(resp_data['chapters']):
            data['chapters'].append(dict(
                slug=chapter['slug_lang_vol_ch_sub'],
                url=chapter['url'],
                title=chapter['full_title'],
                num=chapter['chapter'],
                num_volume=chapter['volume'],
                scanlators=[team['name'] for team in chapter['teams'] if team],
                date=convert_date_string(chapter['published_on'].split('T')[0], format='%Y-%m-%d'),
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        r = self.session_get(self.base_url + '/api' + chapter_url)
        if r.status_code != 200:
            return None

        data = dict(
            pages=[],
        )
        for url in r.json()['chapter']['pages']:
            data['pages'].append(dict(
                slug=None,
                image=url,
            ))

        return data

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(page['image'])
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('?')[0].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return f'{self.base_url}/comics/{slug}'

    def get_latest_updates(self):
        r = self.session_get(self.base_url + '/api/comics')
        if r.status_code != 200:
            return None

        resp_data = r.json()

        results = []
        for item in sorted(resp_data['comics'], key=lambda m: m['updated_at'], reverse=True):
            results.append(dict(
                slug=item['slug'],
                name=item['title'],
                cover=item['thumbnail_small'],
                last_chapter=item['last_chapter']['full_title'] if item['last_chapter'] else None,
            ))

        return results

    def get_most_populars(self):
        r = self.session_get(self.base_url + '/api/comics')
        if r.status_code != 200:
            return None

        resp_data = r.json()

        results = []
        for item in sorted(resp_data['comics'], key=lambda m: m['views'], reverse=True):
            results.append(dict(
                slug=item['slug'],
                name=item['title'],
                cover=item['thumbnail_small'],
                last_chapter=item['last_chapter']['full_title'] if item['last_chapter'] else None,
            ))

        return results

    def search(self, term):
        r = self.session_get(self.base_url + '/api/search/' + term)
        if r.status_code != 200:
            return None

        resp_data = r.json()

        results = []
        for item in resp_data['comics']:
            results.append(dict(
                slug=item['slug'],
                name=item['title'],
                cover=item['thumbnail_small'],
                last_chapter=item['last_chapter']['full_title'] if item['last_chapter'] else None,
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara2


class Webtoonhatti(Madara2):
    id = 'webtoonhatti'
    name = 'Webtoon Hatti'
    lang = 'tr'
    is_nsfw = True

    has_cf = True

    date_format = None
    series_name = 'webtoon'

    base_url = 'https://webtoonhatti.club'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara2


class Argosscan(Madara2):
    id = 'argosscan'
    name = 'Argos Comics'
    lang = 'pt'
    status = 'disabled'  # 2025/11 Switch to a custom website built with Next.js

    base_url = 'https://argoscomic.com'
    logo_url = base_url + '/wp-content/uploads/2024/06/cropped-argos-1-32x32.webp'
    chapters_url = base_url + '/manga/{0}/ajax/chapters/'

    details_synopsis_selector = '.manga-excerpt'
    chapters_selector = '.wp-manga-chapter'



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

# Manga Strean/Manga Reader/Mamga Themesia - WordPress Themes for read manga

# Supported servers:
# Asura Scans For Free [EN] (Disabled)
# Cartel De Manhwas [ES]
# Hijala [AR]
# Iris Scanlator [pt_BR] (Disabled)
# KomikLovers [ID]
# Lelmanga [FR]
# Manga Koleji [TR]
# Neko Scans [ES]
# Noromax [ID]
# Point Zero Toons [pt_BR]
# Ragna Scan [ES] (Disabled)
# Raiki Scan [ES] (Disabled)
# Raw Manga [JA] (Disabled)
# Rimu Scans [FR] (Disabled)
# Rukav Inari [ES] (Disabled)
# Ryujinmanga [ES]
# Senpai Ediciones [ES] (Disabled)
# ShadowMangas [ES]
# SkyMangas [ES]
# SushiScan [FR]
# Sushiscan (.net) [FR]
# Terco Scans [EN]
# Tres Daos [ES] (Disabled)
# VF Scan [FR] (Disabled)
# Vortex Scans For Free [EN] (Disabled)

from gettext import gettext as _
import json
import re

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.servers.utils import get_soup_element_inner_text
from komikku.utils import get_buffer_mime_type
from komikku.webview import CompleteChallenge


class MangaStream(Server):
    base_url: str
    api_url: str = None
    manga_list_url: str = None
    manga_url: str = None
    chapter_url: str = None

    chapters_order: str = 'desc'
    date_format: str = '%B %d, %Y'
    name_re_sub = str = None  # regexp to clean manga name
    series_name: str = 'manga'
    slug_position: int = -2
    chapter_images_re = r'\"images\":(.*?)}'

    name_selector: str = '.entry-title'
    thumbnail_selector: str = '.thumb img'
    authors_selector: str
    genres_selector: str
    scanlators_selector: str
    status_selector: str
    synopsis_selector: str
    chapter_pages_selector: str = '#readerarea img'

    filters = [
        {
            'key': 'type',
            'type': 'select',
            'name': _('Type'),
            'description': _('Filter by Type'),
            'value_type': 'single',
            'default': '',
            'options': [
                {'key': '', 'name': _('All')},
                {'key': 'manga', 'name': _('Manga')},
                {'key': 'manhwa', 'name': _('Manhwa')},
                {'key': 'manhua', 'name': _('Manhua')},
                {'key': 'comic', 'name': _('Comic')},
            ],
        },
    ]

    ignored_chapters_keywords: list = []
    ignored_pages: list = []

    def __init__(self):
        if self.api_url is None:
            self.api_url = self.base_url + '/wp-admin/admin-ajax.php'

        if self.manga_list_url is None:
            self.manga_list_url = self.base_url + '/' + self.series_name + '/'

        if self.manga_url is None:
            self.manga_url = self.base_url + '/' + self.series_name + '/{0}/'

        if self.chapter_url is None:
            self.chapter_url = self.base_url + '/{chapter_slug}/'

        if self.session is None and not self.has_cf:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    @CompleteChallenge()
    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update(dict(
            authors=[],
            scanlators=[],
            genres=[],
            status=None,
            synopsis=None,
            chapters=[],
            server_id=self.id,
        ))

        def compute_status(label):
            if not label:
                return None

            label = label.strip()

            # Ongoing
            labels = (
                'ongoing',
                'coming soon',
                'mass released',
                'daily release',
                'en curso',  # es
                'en cours',  # fr
                'ativo',  # pt
                'em andamento',  # pt_BR
                'devam ediyor',  # tr
            )
            if any(re.findall('|'.join(labels), label, re.IGNORECASE)):
                return 'ongoing'

            # Complete
            labels = (
                'completed',
                'finalizado',  # es
                'fini',  # fr
                'terminé',  # fr
                'completo',  # pt
                'tamamlandı',  # tr
            )
            if any(re.findall('|'.join(labels), label, re.IGNORECASE)):
                return 'complete'

            # Hiatus
            labels = (
                'hiatus',
                'en pause',  # fr
                'hiato',  # pt
                'bırakıldı',  # tr
            )
            if any(re.findall('|'.join(labels), label, re.IGNORECASE)):
                return 'hiatus'

            # Suspended
            labels = (
                'cancelled',
                'dropped',
                'durduruldu',  # tr
                'ara verildi'  # tr
                'cancelado',  # pt_BR
            )
            if any(re.findall('|'.join(labels), label, re.IGNORECASE)):
                return 'suspended'

            return None

        # Name & cover
        data['name'] = soup.select_one(self.name_selector).text.strip()
        if self.name_re_sub:
            data['name'] = re.sub(self.name_re_sub, '', data['name']).strip()

        if element := soup.select_one(self.thumbnail_selector):
            data['cover'] = element.get('data-src')
            if not data['cover']:
                data['cover'] = element.get('data-lazy-src')
                if not data['cover']:
                    data['cover'] = element.get('src')
            if data['cover'] and not data['cover'].startswith('http'):
                data['cover'] = f'https:{data["cover"]}'  # noqa: E231

        # Details
        if self.authors_selector:
            for element in soup.select(self.authors_selector):
                author = get_soup_element_inner_text(element).strip('-')
                if author and author not in data['authors']:
                    data['authors'].append(author)

        if self.genres_selector:
            for element in soup.select(self.genres_selector):
                genre = element.text.strip()
                if genre and genre not in data['genres']:
                    data['genres'].append(genre)

        if self.scanlators_selector:
            for element in soup.select(self.scanlators_selector):
                if scanlator := get_soup_element_inner_text(element).strip('-'):
                    data['scanlators'].append(scanlator)

        if self.status_selector:
            if element := soup.select_one(self.status_selector):
                data['status'] = compute_status(get_soup_element_inner_text(element))

        if self.synopsis_selector:
            if element := soup.select_one(self.synopsis_selector):
                data['synopsis'] = element.text.strip()

        # Chapters
        data['chapters'] = self.get_manga_chapters_data(soup)

        return data

    def get_manga_chapters_data(self, soup):
        chapters = []

        li_elements = soup.select('#chapterlist ul li')
        if self.chapters_order == 'desc':
            li_elements = reversed(li_elements)

        for li_element in li_elements:
            a_element = li_element.select_one('a')

            slug = a_element.get('href').split('/')[-2]
            ignore = False
            for keyword in self.ignored_chapters_keywords:
                if keyword in slug:
                    ignore = True
                    break
            if ignore:
                continue

            title = ' '.join(li_element.select_one('.chapternum').text.split())
            if date_element := li_element.select_one('.chapterdate'):
                date = convert_date_string(date_element.text.strip(), format=self.date_format)
            else:
                date = None

            chapters.append(dict(
                slug=slug,
                title=title,
                num=li_element.get('data-num').strip(),
                date=date,
            ))

        return chapters

    @CompleteChallenge()
    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(
            self.chapter_url.format(manga_slug=manga_slug, chapter_slug=chapter_slug),
            headers={
                'Referer': self.manga_url.format(manga_slug),
            })
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = dict(
            pages=[],
        )

        img_elements = soup.select(self.chapter_pages_selector)
        if not img_elements:
            # Pages images are loaded via javascript
            for script_element in soup.find_all('script'):
                script = script_element.string
                if script is None:
                    continue

                for line in script.split('\n'):
                    line = line.strip()
                    if line.startswith('ts_reader'):
                        if matches := re.compile(self.chapter_images_re).search(line):
                            for image in json.loads(matches.group(1)):
                                if image.split('/')[-1] in self.ignored_pages:
                                    continue

                                data['pages'].append(dict(
                                    slug=None,
                                    image=image,
                                ))
                            break
        else:
            for img_element in img_elements:
                image = img_element.get('data-src')
                if not image:
                    image = img_element.get('src')
                    if not image.startswith('http'):
                        image = f'https:{image}'  # noqa: E231
                if image.split('/')[-1] in self.ignored_pages:
                    continue

                data['pages'].append(dict(
                    slug=None,
                    image=image,
                ))

        return data

    @CompleteChallenge()
    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        headers = {
            'Referer': self.chapter_url.format(manga_slug=manga_slug, chapter_slug=chapter_slug),
        }
        r = self.session_get(page['image'], headers=headers)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, title=None, type=None, orderby=None):
        r = self.session_get(
            self.manga_list_url,
            params=dict(
                status='',
                type=type,
                order=orderby or '',
                title=title,
            )
        )
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select('.listupd .bs a'):
            name = a_element.get('title')
            if self.name_re_sub:
                name = re.sub(self.name_re_sub, '', name).strip()

            if cover_element := a_element.select_one('img.ts-post-image'):
                if cover_element.get('data-lazy-src'):
                    cover = cover_element.get('data-lazy-src')
                elif cover_element.get('data-src'):
                    cover = cover_element.get('data-src')
                else:
                    cover = cover_element.get('src')
            else:
                continue

            results.append(dict(
                slug=a_element.get('href').split('/')[self.slug_position],
                name=name,
                cover=cover,
            ))

        return results

    @CompleteChallenge()
    def get_latest_updates(self, type):
        return self.get_manga_list(type=type, orderby='update')

    @CompleteChallenge()
    def get_most_populars(self, type):
        return self.get_manga_list(type=type, orderby='popular')

    @CompleteChallenge()
    def search(self, term, type):
        return self.get_manga_list(title=term, type=type)



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import ServerContent


class Annyo(Server):
    id = 'annyo'
    name = 'Annyo - le mouton a 5 pattes'
    lang = 'fr'
    content = ServerContent(
        type=[_('Webcomic'), _('Self-publishing')],
        license='CC BY-SA'
    )
    true_search = False

    base_url = 'https://annyo.logaton.fr'
    logo_url = base_url + '/themes/margot/images/favicon.png'
    chapters_url = base_url + '/?PagePrincipale'
    image_url = base_url + '/{0}'
    cover_url = base_url + '/cache/PageTitre_Annyo_vignette_140_97_20241006122801_20241006142905.jpg'

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'User-Agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns comic data by scraping comic HTML page content
        """
        data = initial_data.copy()
        data.update(dict(
            authors=['Kalam'],
            scanlators=[],
            genres=['Écologie', 'Vulgarisation', 'Humour'],
            status='ongoing',
            synopsis="""« Si le berger l’a dit, c’est que c’est comme ça ». Le chien de berger est d’accord. Et aussi l’oiseau que nourrit le berger. Mais il y a ce mouton vert à 5 pattes qui doit toujours ramener sa fraise, ça complique tout. C’était plus simple de penser qu’il fallait donner sa laine et en vouloir aux moutons noirs. Enfin, il est quand même bizarre : il est vert et il a 5 pattes.
N’y voyez aucune similitude avec des situations réelles. Vraiment aucune.""",
            chapters=[],
            server_id=self.id,
            cover=self.cover_url,
        ))

        # Chapters
        r = self.session_get(self.chapters_url)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        for a_element in reversed(soup.select('.pbgalery-link')):
            date = a_element.get('title')
            data['chapters'].append(dict(
                slug=date.replace('-', ''),
                date=convert_date_string(date, '%Y-%m-%d'),
                title=a_element.h5.text.strip(),
            ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns comic chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.base_url)
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        if a_element := soup.select_one(f'[data-id_fiche="{chapter_slug}"] a.pbgalery-link'):
            return dict(
                pages=[
                    dict(
                        slug=None,
                        image=a_element.get('href'),
                    ),
                ]
            )

        return None

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(self.image_url.format(page['image']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['image'].split('/')[-1],
        )

    def get_manga_url(self, slug, url):
        """
        Returns comic absolute URL
        """
        return self.base_url

    def get_most_populars(self):
        return [dict(
            slug='',
            name='Annyo - le mouton a 5 pattes',
            cover=self.cover_url,
        )]

    def search(self, term=None):
        # This server does not have a search
        # but a search method is needed for `Global Search` in `Explorer`
        # In order not to be offered in `Explorer`, class attribute `true_search` must be set to False

        results = []
        for item in self.get_most_populars():
            if term and term.lower() in item['name'].lower():
                results.append(item)

        return results



# SPDX-FileCopyrightText: 2019-2026 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from gettext import gettext as _
import time

from bs4 import BeautifulSoup
import requests

from komikku.consts import DOWNLOAD_MAX_DELAY
from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.servers.utils import convert_date_string
from komikku.utils import get_buffer_mime_type
from komikku.utils import get_response_elapsed
from komikku.utils import is_number


class Rncalation(Server):
    id = 'rncalation'
    name = 'Rncalation'
    lang = 'es'

    base_url = 'https://rncalation.online'
    logo_url = base_url + '/icons/icon-96.png'
    search_url = base_url + '/library'
    manga_url = base_url + '/comics/{0}'
    chapters_url = base_url + '/comics/{0}/chapters'
    chapter_url = base_url + '/leer/{0}'

    filters = [
        {
            'key': 'type',
            'type': 'select',
            'name': _('Type'),
            'description': _('Filter by Type'),
            'value_type': 'single',
            'default': '',
            'options': [
                {'key': '', 'name': _('Any')},
                {'key': 'Manga', 'name': _('Manga')},
                {'key': 'Manhwa', 'name': _('Manhwa')},
                {'key': 'Manhua', 'name': _('Manhua')},
                {'key': 'Doujinshi', 'name': _('Doujinshi')},
                {'key': 'Other', 'name': _('Other')},
            ]
        },
        {
            'key': 'status',
            'type': 'select',
            'name': _('Status'),
            'description': _('Filter by Status'),
            'value_type': 'single',
            'default': '',
            'options': [
                {'key': '', 'name': _('Any')},
                {'key': 'Ongoing', 'name': _('Ongoing')},
                {'key': 'Completed', 'name': _('Completed')},
                {'key': 'Hiatus', 'name': _('Hiatus')},
                {'key': 'Cancelled', 'name': _('Canceled')},
            ]
        },
    ]

    headers = {
        'User-Agent': USER_AGENT,
    }

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers = self.headers

    def get_manga_data(self, initial_data):
        """
        Returns manga data by scraping manga HTML page content

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Manga slug is missing in initial data'

        r = self.session_get(self.manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if mime_type != 'text/html':
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = initial_data.copy()
        data.update({
            'authors': [],
            'scanlators': [],
            'genres': [],
            'status': None,
            'synopsis': None,
            'chapters': [],
            'server_id': self.id,
            'cover': None,
        })

        data['name'] = soup.select_one('h1').text.strip()
        data['cover'] = self.base_url + soup.select_one('img.comic-cover__img').get('src')

        # Status
        if element := soup.select_one('.flex.items-baseline:-soup-contains("Estado") span.text-right'):
            status = element.text.strip()
            if status == 'Completado':
                data['status'] = 'complete'
            elif status == 'En emisión':
                data['status'] = 'ongoing'
            elif status == 'En pausa':
                data['status'] = 'hiatus'
            elif status == 'Cancelado':
                data['status'] = 'suspended'

        # Authors
        if element := soup.select_one('.flex.items-baseline:-soup-contains("Autor") span.text-right'):
            data['authors'].append(element.text.strip())

        if element := soup.select_one('.flex.items-baseline:-soup-contains("Arte") span.text-right'):
            artist = element.text.strip()
            if artist not in data['authors']:
                data['authors'].append(artist)

        # Genres
        for element in soup.select('span.inline-flex.items-center.rounded'):
            badge = element.text.strip().lower().capitalize()
            if badge in ('Manga', 'Manhwa', 'Manhua', 'Doujinshi'):
                data['genres'].append(badge)
                break

        if element := soup.select_one('.flex.items-baseline:-soup-contains("Demográfico") span.text-right'):
            demogrpahic = element.text.strip()
            if demogrpahic not in data['genres']:
                data['genres'].append(demogrpahic)

        if element := soup.select_one('.flex.items-baseline:-soup-contains("Géneros") span.text-right'):
            for genre in element.text.split(','):
                data['genres'].append(genre.strip())

        # Scanlators
        if element := soup.select_one('.flex.flex-col > a'):
            data['scanlators'].append(element.text.strip())

        # Synopsis
        data['synopsis'] = soup.select_one('p.m-0').text.strip()

        # Chapters
        data['chapters'] = self.get_manga_chapters_data(data['slug'])

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        """
        Returns manga chapter data by scraping chapter HTML page content

        Currently, only pages are expected.
        """
        r = self.session_get(self.chapter_url.format(chapter_slug))
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        data = {
            'pages': [],
        }
        for img_element in soup.select('img.page-img, .page-wrap img'):
            image = img_element.get('data-src') or img_element.get('src')
            if not image:
                continue

            data['pages'].append({
                'image': image,
                'slug': None,
            })

        return data

    def get_manga_chapters_data(self, manga_slug):
        def get_page(page):
            r = self.session_get(
                self.chapters_url.format(manga_slug),
                params={
                    'page': page,
                }
            )
            if r.status_code != 200:
                return None

            soup = BeautifulSoup(r.text, 'lxml')

            items = []
            for a_element in soup.select('a'):
                if date_element := a_element.select_one('span.uppercase'):
                    date = date_element.text.strip()
                else:
                    date = None
                num = a_element.get('data-chapter-num')

                items.append({
                    'title': a_element.get('data-chapter-label'),
                    'slug': a_element.get('href').split('/')[-1],
                    'num': num if is_number(num) else None,
                    'date': convert_date_string(date, format='%B %d, %Y'),
                })

            more = len(items) == 30

            return items, more, get_response_elapsed(r)

        chapters = []
        delay = None
        more = True
        page = 1
        while more:
            if delay:
                time.sleep(delay)

            items, more, rtime = get_page(page)
            chapters += items

            delay = min(rtime * 2, DOWNLOAD_MAX_DELAY) if rtime else None
            page += 1

        return list(reversed(chapters))

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(
            page['image'],
            headers={
                'Referer': f'{self.base_url}/',
            }
        )
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return {
            'buffer': r.content,
            'mime_type': mime_type,
            'name': page['image'].split('/')[-1],
        }

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_manga_list(self, term=None, type=None, status=None, orderby=None):
        params = {
            'type': type,
            'status': status,
        }
        if term is not None:
            params['q'] = term
            params['sort'] = 'title'
        elif orderby == 'populars':
            params['sort'] = 'views'
        elif orderby == 'latest':
            params['sort'] = 'latest'

        r = self.session_get(self.search_url, params=params)
        if r.status_code != 200:
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = []
        for a_element in soup.select('.lib-grid > a.comic-card'):
            type_ = a_element.select_one('.absolute.top-2.left-2').text.strip()
            if type_ == 'Novel':
                continue

            img_element = a_element.select_one('img.comic-cover__img')
            results.append({
                'slug': a_element.get('href').split('/')[-1],
                'name': img_element.get('alt'),
                'cover': self.base_url + img_element.get('src'),
            })

        return results

    def get_latest_updates(self, type=None, status=None):
        """
        Returns latest updates
        """
        return self.get_manga_list(type=type, status=status, orderby='latest')

    def get_most_populars(self, type=None, status=None):
        """
        Returns most viewed
        """
        return self.get_manga_list(type=type, status=status, orderby='populars')

    def search(self, term, type=None, status=None):
        return self.get_manga_list(term=term, type=type, status=status)



# SPDX-FileCopyrightText: 2021-2024 Lili Kurek
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Lili Kurek <lilikurek@proton.me>

# Supported servers:
# Danke fürs Lesen [EN]
# Hachirumi [EN]
# Magical Translators [EN/PL]

import datetime

from bs4 import BeautifulSoup
import requests

from komikku.consts import USER_AGENT
from komikku.servers import Server
from komikku.utils import get_buffer_mime_type


class Guya(Server):
    base_url: str
    manga_url: str
    api_manga_url: str
    page_image_url: str

    def __init__(self):
        if self.session is None:
            self.session = requests.Session()
            self.session.headers.update({'user-agent': USER_AGENT})

    def get_manga_data(self, initial_data):
        """
        Returns manga data from API

        Initial data should contain at least manga's slug (provided by search)
        """
        assert 'slug' in initial_data, 'Slug is missing in initial data'

        r = self.session_get(self.api_manga_url.format(initial_data['slug']))
        if r.status_code != 200:
            return None

        resp_data = r.json()

        data = initial_data.copy()
        data.update(dict(
            authors=[resp_data['author']],
            scanlators=list(resp_data['groups'].values()),
            genres=[],
            status=None,
            cover=self.base_url + resp_data['cover'],
            synopsis=BeautifulSoup(resp_data['description'], 'lxml').text.strip() if resp_data['description'] else None,
            chapters=self.resolve_chapters(initial_data['slug']),
            server_id=self.id,
        ))

        return data

    def get_manga_chapter_data(self, manga_slug, manga_name, chapter_slug, chapter_url):
        chapters = self.resolve_chapters(manga_slug)
        if not chapters:
            return None

        for chapter in chapters:
            if chapter['slug'] == chapter_slug:
                return chapter

        return None

    def get_manga_chapter_page_image(self, manga_slug, manga_name, chapter_slug, page):
        """
        Returns chapter page scan (image) content
        """
        r = self.session_get(self.page_image_url.format(manga_slug, chapter_slug, page['slug']))
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('image'):
            return None

        return dict(
            buffer=r.content,
            mime_type=mime_type,
            name=page['slug'].split('_')[0],
        )

    def get_manga_url(self, slug, url):
        """
        Returns manga absolute URL
        """
        return self.manga_url.format(slug)

    def get_latest_updates(self):
        r = self.session_get(self.base_url + '/latest_chapters/')
        if r.status_code != 200:
            return None

        mime_type = get_buffer_mime_type(r.content)
        if not mime_type.startswith('text/html'):
            return None

        soup = BeautifulSoup(r.text, 'lxml')

        results = dict()
        for a_element in soup.select('#chapterTable > tr > td:nth-child(2) > a'):
            slug = a_element.get('href').split('/')[-2]
            if slug not in results:
                results[slug] = dict(
                    slug=slug,
                    name=a_element.text.strip(),
                )

        return results.values()

    def get_most_populars(self):
        return self.search('')

    def resolve_chapters(self, manga_slug):
        data = []

        r = self.session_get(self.api_manga_url.format(manga_slug))
        if r.status_code != 200:
            return None

        resp_data = r.json()
        groups = resp_data['groups']
        chapters = resp_data['chapters']

        for chapter_id, chapter in chapters.items():
            title = '#' + chapter_id
            if chapter['title']:
                title += ' - ' + chapter['title']

            for group_id, pages in chapter['groups'].items():
                data.append(dict(
                    slug=chapter['folder'] + '/' + group_id,
                    title=title,
                    num=chapter_id,
                    num_volume=chapter['volume'],
                    pages=[dict(slug=slug, image=None) for slug in pages],
                    scanlators=[groups[group_id]],
                    date=datetime.date.fromtimestamp(chapter['release_date'][group_id]),
                ))

        return data

    def search(self, term):
        params = None
        r = self.session_get(self.base_url + '/api/get_all_series/', params=params)
        if r.status_code != 200:
            return None

        resp_data = r.json()
        results = []

        for name, item in resp_data.items():
            if term and term.casefold() not in name.casefold():
                continue

            # The only indicator of a series being in Polish on Magical Translators is slug ending with PL.
            if self.id == 'magicaltranslators':
                if item['slug'].endswith('-PL'):
                    continue
            elif self.id == 'magicaltranslators_pl':
                if not item['slug'].endswith('-PL'):
                    continue

            results.append(dict(
                slug=item['slug'],
                name=name,
                cover=self.base_url + item['cover'],
            ))

        return results



# SPDX-FileCopyrightText: 2019-2025 Valéry Febvre
# SPDX-License-Identifier: GPL-3.0-or-later
# Author: Valéry Febvre <vfebvre@easter-eggs.com>

from komikku.servers.multi.madara import Madara


class Mangazin(Madara):
    id = 'mangazin'
    name = 'MangaZin'
    lang = 'en'
    is_nsfw = True

    base_url = 'https://mangazin.org'
    logo_url = base_url + '/wp-content/uploads/2020/02/cropped-IMG_20200222_225535-32x32.jpg'
